"""Definitions of Dash layout."""

import os.path

import dash_bootstrap_components as dbc
import dash_core_components as dcc
import dash_html_components as html
import plotly.express as px
from matplotlib.colors import rgb2hex

import seaborn as sns

from .cache import cache
from . import settings
from . import store
from . import ui_plots
from ..__init__ import __version__
from .settings import reassignSettings


def render_navbar():
    """Render the site navbar"""
    return dbc.Navbar(
        dbc.Container(
            children=[
                # Use row and col to control vertical alignment of logo / brand
                dbc.NavbarBrand(
                    [html.I(className="fas fa-home mr-1"), settings.HOME_BRAND], id="page-brand"
                ),
            ]
        ),
        # className="mb-5",
        color="primary",
        dark=True,
        id="page-navbar",
    )


@cache.memoize()
def render_pca(us: reassignSettings):
    return px.scatter(
        store.compute_pca(us).sort_values(by="panel"),
        x="X",
        y="Y",
        color="panel",
        color_discrete_sequence=[rgb2hex(v) for v in store.compute_clustercoldict(us).values()],
    )


@cache.memoize()
def render_tsne(us: reassignSettings):
    return px.scatter(
        store.compute_tsne(us).sort_values(by="panel"),
        x="X",
        y="Y",
        color="panel",
        color_discrete_sequence=[rgb2hex(v) for v in store.compute_clustercoldict(us).values()],
    )


@cache.memoize()
def render_agg_clust(us: reassignSettings):
    return px.scatter(
        store.compute_acluster(us)[0].sort_values(by="new_assignments"),
        x="X",
        y="Y",
        color="new_assignments",
        color_discrete_sequence=[rgb2hex(v) for v in store.compute_clustercoldict(us).values()],
    )


@cache.memoize()
def render_image_cluster_panels(us: reassignSettings):
    data = store.load_all_data(us)
    img_b64 = ui_plots.plot_clustermap_panels_as_base64(data, store.compute_panelcoldict(us))
    return html.Img(src="data:image/png;base64,%s" % img_b64, className="img-responsive")


@cache.memoize()
def render_image_cluster_clustering(us: reassignSettings):
    data = store.load_all_data(us)
    img_b64 = ui_plots.plot_clustermap_clustering_as_base64(
        data, *store.compute_acluster(us), store.compute_clustercoldict(us)
    )
    return html.Img(src="data:image/png;base64,%s" % img_b64, className="img-responsive")


# @cache.memoize()
# def render_image_batches_clustering(us: reassignSettings):
#    data = store.load_all_data(us)
#    img_b64 = ui_plots.plot_clustermap_batches_as_base64(
#        data, store.compute_acluster(us), store.compute_clustercoldict(us)
#    )
#    return html.Img(src="data:image/png;base64,%s" % img_b64, className="img-responsive")


@cache.memoize()
def render_images_batch_separation(us: reassignSettings):
    return [
        dcc.Graph(figure=px.scatter(p, x="X", y="Y", color="batch"), id="batch-separation-%i" % i)
        for (i, p) in enumerate(store.compute_batches(us))
    ]


def render_main_content():
    """Render page main content"""

    # thresh: int = 50
    # pca_components: int = 20
    # batch_num: float = 0.985
    # pca_seed: int = 100

    controls = dbc.Card(
        children=[
            html.P("Clustering Parameters", className="font-weight-bold"),
            dbc.FormGroup(
                children=[
                    dbc.Label("Threshold"),
                    dbc.Input(
                        id="input-thresh",
                        type="number",
                        value=settings.reassign_SETTINGS.threshold,
                        debounce=True,
                    ),
                ],
            ),
            dbc.FormGroup(
                children=[
                    dbc.Label("PCA Components"),
                    dbc.Input(
                        id="input-pca-components",
                        type="number",
                        value=settings.reassign_SETTINGS.pca_components,
                        debounce=True,
                    ),
                ],
            ),
            dbc.FormGroup(
                children=[
                    dbc.Label("number of batches (clusters)"),
                    dbc.Input(
                        id="input-batch-factor",
                        type="text",
                        value=settings.reassign_SETTINGS.batch_num,
                        debounce=True,
                    ),
                ],
            ),
            html.Div(id="buffer-cluster-params", style={"display": "none"}),
            dbc.Button("Save", id="input-save-batches-button"),
            html.P(id="buffer-save-batches-text"),
        ],
        body=True,
    )

    def make_card(id_, type_="graph"):
        if type_ == "graph":
            return dbc.Card(
                dbc.CardBody(
                    dbc.Spinner([dcc.Graph(id=id_)], color="primary"),
                    className="mt-3 text-center",
                ),
                className="border-top-0 rounded-0",
            )
        else:
            return dbc.Card(
                dbc.CardBody(
                    dbc.Spinner(color="primary", id=id_),
                    className="mt-3 text-center",
                ),
                className="border-top-0 rounded-0",
            )

    tabs_content = [
        dbc.Tab(make_card("graph-pca"), label="PCA"),
        dbc.Tab(make_card("graph-tsne"), label="tSNE"),
        dbc.Tab(make_card("graph-agg-clust"), label="Agg. Clust."),
        dbc.Tab(
            make_card("container-image-cluster-panels", type_="html"), label="Original assignment"
        ),
        dbc.Tab(
            make_card("container-image-cluster-clustering", type_="html"), label="New assignment"
        ),
        # dbc.Tab(
        #    make_card("container-image-batch-clustering", type_="html"), label="Batches clustering"
        # ),
        dbc.Tab(make_card("container-image-batch-separation", type_="html"), label="Batches"),
    ]

    return html.Div(
        children=[
            dbc.Row(
                children=[
                    dbc.Col(
                        children=[controls],
                        id="controls",
                        md=3,
                    ),
                    dbc.Col(children=[dbc.Tabs(tabs_content)], id="plots", md=9),
                ]
            )
        ],
        className="container pt-3",
    )


def render_footer():
    """Render page footer"""
    return html.Footer(
        html.Div(
            children=[
                html.Div(
                    children=[
                        html.Div(
                            children=[
                                html.Span(
                                    "clear-CNV v%s by BIH CUBI" % __version__,
                                    className="text-muted",
                                )
                            ],
                            className="col-5",
                        ),
                        html.Div(
                            children=[
                                html.A(
                                    children=[
                                        html.I(className="fas fa-book mr-1"),
                                        "Manual & Tutorial",
                                    ],
                                    href="https://clear-cnv.readthedocs.io",
                                    className="text-muted mr-3",
                                ),
                                html.A(
                                    children=[
                                        html.I(className="fas fa-globe-europe mr-1"),
                                        "CUBI Homepage",
                                    ],
                                    href="https://www.cubi.bihealth.org",
                                    className="text-muted mr-3",
                                ),
                                html.A(
                                    children=[
                                        html.I(className="fab fa-github mr-1"),
                                        "GitHub Project",
                                    ],
                                    href="https://github.com/bihealth/clear-cnv",
                                    className="text-muted",
                                ),
                            ],
                            className="col-7 text-right",
                        ),
                    ],
                    className="row",
                )
            ],
            className="container",
        ),
        className="footer",
    )


def build_layout():
    """Build the overall Dash app layout"""
    return html.Div(
        children=[
            # Represents the URL bar, doesn't render anything.
            dcc.Location(id="url", refresh=False),
            # Navbar, content, footer.
            render_navbar(),
            render_main_content(),
            render_footer(),
        ],
        id="_dash-app-content",  # make pytest-dash happy
    )



#!/usr/bin/env python

import argparse
import os
import tempfile

from logzero import logger

from . import misc
from . import matchscores
from . import cnv_calling
from . import visualize_scores
from . import workflow_cnv_calling
from . import workflow_reassignment
from . import __version__

#: The executables required for running clear-CNV.
REQUIRED_EXECUTABLES = ("bedops", "bedtools", "sort")


def get_parser():
    """Return argparse command line parser."""
    parser = argparse.ArgumentParser(
        description="clearCNV can compute matchscores, CNV-calls and visualizations.",
    )
    parser.add_argument("--version", action="version", version=__version__)
    subparsers = parser.add_subparsers()

    # =========================================================================
    #  matchscores
    # =========================================================================
    parser_matchscores = subparsers.add_parser(
        "matchscores", description="Matchscore calculation script."
    )
    parser_matchscores.add_argument(
        "-p", "--panel", help="Name of the data set (or panel)", required=True, type=str
    )
    parser_matchscores.add_argument(
        "-c",
        "--coverages",
        help="Coverages file in tsv format",
        required=True,
        type=os.path.abspath,
    )
    parser_matchscores.add_argument(
        "-m",
        "--matchscores",
        help="Output matchscores.tsv file",
        required=True,
        type=os.path.abspath,
    )
    parser_matchscores.add_argument(
        "-x",
        "--expected_artefacts",
        help="Expected ratio of CNVs or artefacs in target fragment counts",
        required=False,
        type=float,
        default=0.02,
    )
    parser_matchscores.add_argument(
        "--cores",
        help="Number of cpu cores used in parallel processing. Default: determined automatically.",
        required=False,
        type=int,
        default=0,
    )
    parser_matchscores.add_argument(
        "--fast",
        help="If set, clearCNV will speed up matchscore calculation by taking only at most 2.000 targets per sample into account.",
        required=False,
        dest="fast",
        action="store_true",
    )
    parser_matchscores.set_defaults(fast=False, func=matchscores.matchscores)

    # =========================================================================
    #  cnv calling
    # =========================================================================
    parser_cnv_calling = subparsers.add_parser(
        "cnv_calling",
        description=(
            "CNV calling script. Output is a single file in tsv format containing a list of CNV "
            "calls sorted by score. Some quality control plots are added to the analysis "
            "directory in the process."
        ),
    )
    parser_cnv_calling.add_argument(
        "-p", "--panel", help="Name of the data set(or panel)", required=True, type=str
    )
    parser_cnv_calling.add_argument(
        "-c",
        "--coverages",
        help="Coverages file in tsv format",
        required=True,
        type=os.path.abspath,
    )
    parser_cnv_calling.add_argument(
        "-a",
        "--analysis_directory",
        help="Path to the directory, where analysis files are stored",
        required=True,
        type=os.path.abspath,
    )
    parser_cnv_calling.add_argument(
        "-m",
        "--matchscores",
        help="matchscores.tsv file generated with matchscores.py",
        required=True,
        type=os.path.abspath,
    )
    parser_cnv_calling.add_argument(
        "-C",
        "--cnv_calls",
        help="Output cnv_calls.tsv file formatted in tsv format",
        required=True,
        type=os.path.abspath,
    )
    parser_cnv_calling.add_argument(
        "-r",
        "--ratio_scores",
        help="Output ratio scores file in tsv format. Best kept together with cnv_calls.tsv",
        required=True,
        type=os.path.abspath,
    )
    parser_cnv_calling.add_argument(
        "-z",
        "--z_scores",
        help="Output z-scores file in tsv format. Best kept together with cnv_calls.tsv",
        required=True,
        type=os.path.abspath,
    )
    parser_cnv_calling.add_argument(
        "-x",
        "--expected_artefacts",
        help="Expected ratio of CNVs or artefacs in target fragment counts",
        required=False,
        type=float,
        default=0.02,
    )
    parser_cnv_calling.add_argument(
        "-u",
        "--sample_score_factor",
        help="The factor u multiplied with the median sample score to define sample groups. u should range between 1.0 < u < 5.0. Default is 2.0.",
        required=False,
        type=float,
        default=2.0,
    )
    parser_cnv_calling.add_argument(
        "-g",
        "--minimum_group_sizes",
        help="Group size per CNV calling group per match scores. Default is 30.",
        required=False,
        type=int,
        default=30,
    )
    parser_cnv_calling.add_argument(
        "-s",
        "--zscale",
        help="A higher z-scale results in more CNV calls. Should only be 0.0 <= zscale <= 2.0. Default is 0.65.",
        required=False,
        type=float,
        default=0.65,
    )
    parser_cnv_calling.add_argument(
        "--del_cutoff",
        help="A hard threshold on the ratio score for deletions. Default is 0.75.",
        required=False,
        type=float,
        default=0.75,
    )
    parser_cnv_calling.add_argument(
        "--dup_cutoff",
        help="A hard threshold on the ratio score for duplications. Default is 1.35.",
        required=False,
        type=float,
        default=1.35,
    )
    parser_cnv_calling.add_argument(
        "--trans_prob",
        help="Transition probability of the HMM to change state from WT to CNV. Default is 0.001. Lower values prefer longer CNVs, higher values prefer shorter CNVs.",
        required=False,
        type=float,
        default=0.001,
    )
    parser_cnv_calling.add_argument(
        "--plot_regions",
        dest="plot_regions",
        help="If set, the CNV calling script plots heatmaps of the CNV called regions with the corresponding sample group taken as background.",
        required=False,
        action="store_true",
    )
    parser_cnv_calling.set_defaults(plot_regions=False)
    # parser_cnv_calling.add_argument(
    #    "--plot_regions",
    #    help="If true, the CNV calling script plots heatmaps of the regions.",
    #    required=False,
    #    type=bool,
    #    default=False,
    # )
    parser_cnv_calling.add_argument(
        "--cores",
        help="Number of cpu cores used in parallel processing. Default: determined automatically.",
        required=False,
        type=int,
        default=0,
    )
    parser_cnv_calling.set_defaults(func=cnv_calling.cnv_calling)

    # =========================================================================
    #  visualize
    # =========================================================================
    parser_visualize = subparsers.add_parser(
        "visualize",
        description=(
            "The visualization script creates html files containing heatmap-like matrices "
            "containing the ratios aligned with mappability, GC-content and target size so "
            "that CNVs can be visually identified and evaluated easily."
        ),
    )
    parser_visualize.add_argument(
        "-a",
        "--analysis_directory",
        help="Path to the directory, where analysis files are stored",
        required=True,
        type=os.path.abspath,
    )
    parser_visualize.add_argument(
        "-r",
        "--ratio_scores",
        help="Ratio scores file in tsv format, generated in cnv_calling.py",
        required=True,
        type=os.path.abspath,
    )
    parser_visualize.add_argument(
        "-z",
        "--z_scores",
        help="Z-scores file in tsv format, generated in cnv_calling.py",
        required=True,
        type=os.path.abspath,
    )
    parser_visualize.add_argument(
        "-n",
        "--annotated",
        help="BED-formatted annotatins file generated with clearCNV annotations. E.g. annotations.bed.",
        required=True,
        type=os.path.abspath,
    )
    parser_visualize.add_argument(
        "-s",
        "--size",
        help="Rough number of targets in each visualization. Defaults at 1000.",
        required=False,
        type=int,
        default=1000,
    )
    parser_visualize.set_defaults(func=visualize_scores.visualize)

    # =========================================================================
    #  merge bed
    # =========================================================================
    parser_merge_bed = subparsers.add_parser(
        "merge_bed",
        description=("Merges bed files to non-overlapping intervals."),
    )
    parser_merge_bed.add_argument(
        "-i",
        "--infile",
        help="Path to the original .bed file.",
        required=True,
        type=os.path.abspath,
    )
    parser_merge_bed.add_argument(
        "-o",
        "--outfile",
        help="Output path to the merged .bed file.",
        required=True,
        type=os.path.abspath,
    )
    parser_merge_bed.set_defaults(func=misc.merge_bedfile)

    # =========================================================================
    #  reassignment
    # =========================================================================

    # prepare - maybe exclude in future
    parser_prepare_reassign = subparsers.add_parser(
        "prepare_reassignment",
        description=(
            "Prepares the necessary files to perform panel reassignment on a set of .bed files and corresponding .bam file data sets."
        ),
    )
    parser_prepare_reassign.add_argument(
        "-m",
        "--metafile",
        help="Path to the file containing the meta information. It is a .tsv of the scheme -panel bams.txt bed.bed. \
            It aligns each desired panel name with the corresponding .bam files and the .bed file.",
        required=True,
        type=os.path.abspath,
    )
    parser_prepare_reassign.add_argument(
        "-b",
        "--bamsfile",
        help="Output .txt file. It will contain the distinct set of all given .bam file paths.",
        required=True,
        type=os.path.abspath,
    )
    parser_prepare_reassign.add_argument(
        "-d",
        "--bedfile",
        help="Output .bed file. It will contain the merged union of all given .bed files.",
        required=True,
        type=os.path.abspath,
    )
    parser_prepare_reassign.set_defaults(func=misc.prepare_reassignment)

    # =========================================================================
    #  coverage
    # =========================================================================
    parser_rtbeds = subparsers.add_parser(
        "coverage",
        description=(
            "wrapper for bedtools multicov. Creates an .rtbed file which contains the read depth coverage per target."
        ),
    )
    parser_rtbeds.add_argument(
        "-i",
        "--inbam",
        help="Path to the .bam file of the sample.",
        required=True,
        type=os.path.abspath,
    )
    parser_rtbeds.add_argument(
        "-b",
        "--bedfile",
        help="Path to the merged .bed file.",
        required=True,
        type=os.path.abspath,
    )
    parser_rtbeds.add_argument(
        "-r",
        "--rtbed",
        help="Output file in .rtbed format.",
        required=True,
        type=os.path.abspath,
    )
    parser_rtbeds.set_defaults(func=misc.count_pair)

    # -------------------------------------------------------------------------
    # merge rtbeds

    parser_merge_rtbeds = subparsers.add_parser(
        "merge_coverages",
        description=("Merges all .rtbed files into one table in .tsv format."),
    )
    parser_merge_rtbeds.add_argument(
        "-b",
        "--bedfile",
        help="Path to the merged .bed file.",
        required=True,
        type=os.path.abspath,
    )
    parser_merge_rtbeds.add_argument(
        "-r",
        "--rtbeds",
        help="Input .rtbed file paths.",
        required=True,
        nargs="+",
        type=os.path.abspath,
    )
    parser_merge_rtbeds.add_argument(
        "-c",
        "--coverages",
        help="Output table in .tsv format.",
        required=True,
        type=os.path.abspath,
    )
    parser_merge_rtbeds.set_defaults(func=misc.merge_rtbeds)

    # =========================================================================
    #  annotations
    # =========================================================================
    parser_annotations = subparsers.add_parser(
        "annotations",
        description=("Creates annotations file."),
    )
    parser_annotations.add_argument(
        "-r",
        "--reference",
        help="Path to the genomic reference.",
        required=True,
        type=os.path.abspath,
    )
    parser_annotations.add_argument(
        "-b",
        "--bedfile",
        help="Path to the merged .bed file.",
        required=True,
        type=os.path.abspath,
    )
    parser_annotations.add_argument(
        "-k",
        "--kmer_align",
        help="Path to aligned k-mers (mappability) file in .bed format.",
        required=True,
        type=os.path.abspath,
    )
    parser_annotations.add_argument(
        "-a",
        "--annotations",
        help="Output file in .bed format.",
        required=True,
        type=os.path.abspath,
    )
    parser_annotations.set_defaults(func=misc.create_annotations_file)

    # =========================================================================
    #  workflow CNV calling
    # =========================================================================
    parser_workflow_cnv_calling = subparsers.add_parser(
        "workflow_cnv_calling",
        description=(
            "Complete CNV-calling snakemake-workflow to run on data from a single sequencing panel (bed-file)."
        ),
    )
    parser_workflow_cnv_calling.add_argument(
        "-w",
        "--workdir",
        help="Path to the snakemake workdir.",
        required=True,
        type=os.path.abspath,
    )
    parser_workflow_cnv_calling.add_argument(
        "-p",
        "--panelname",
        help="name of the panel or dataset.",
        required=True,
        type=str,
    )
    parser_workflow_cnv_calling.add_argument(
        "-r",
        "--reference",
        help="Path to the genomic reference.",
        required=True,
        type=os.path.abspath,
    )
    parser_workflow_cnv_calling.add_argument(
        "-b",
        "--bamsfile",
        help="Path to a .txt file that contains all paths to all used .bam files.",
        required=True,
        type=os.path.abspath,
    )
    parser_workflow_cnv_calling.add_argument(
        "-d",
        "--bedfile",
        help="Path to the BED file that contains all targets of the sequencing panel. It must contain the columns: chr, start, end, gene.",
        required=True,
        type=os.path.abspath,
    )
    parser_workflow_cnv_calling.add_argument(
        "-k",
        "--kmer_align",
        help="Path to aligned k-mers (mappability) file in .bed format.",
        required=True,
        type=os.path.abspath,
    )
    parser_workflow_cnv_calling.add_argument(
        "-c",
        "--cores",
        help="Number of used cores in snakemake workflow. Default is 32.",
        required=False,
        type=int,
        default=32,
    )
    # options
    parser_workflow_cnv_calling.add_argument(
        "--expected_artefacts",
        help="Expected ratio of CNVs or artefacs in target fragment counts. Default is 0.02",
        required=False,
        type=float,
        default=0.02,
    )
    parser_workflow_cnv_calling.add_argument(
        "--sample_score_factor",
        help="The factor u multiplied with the median sample score to define sample groups. u should range between 1.0 < u < 5.0. Default is 2.0.",
        required=False,
        type=float,
        default=2.0,
    )
    parser_workflow_cnv_calling.add_argument(
        "--minimum_group_sizes",
        help="Minimum group size per CNV calling group per match scores. Default is 20.",
        required=False,
        type=int,
        default=20,
    )
    parser_workflow_cnv_calling.add_argument(
        "--zscale",
        help="A higher z-scale results in more CNV calls. Should only be 0.0 <= zscale <= 2.0. Default is 0.65.",
        required=False,
        type=float,
        default=0.65,
    )
    parser_workflow_cnv_calling.add_argument(
        "--size",
        help="Rough number of targets in each visualization. Default is 2000.",
        required=False,
        type=int,
        default=2000,
    )
    parser_workflow_cnv_calling.add_argument(
        "--del_cutoff",
        help="A hard threshold on the ratio score for deletions. Default is 0.75.",
        required=False,
        type=float,
        default=0.75,
    )
    parser_workflow_cnv_calling.add_argument(
        "--dup_cutoff",
        help="A hard threshold on the ratio score for duplications. Default is 1.35.",
        required=False,
        type=float,
        default=1.35,
    )
    parser_workflow_cnv_calling.add_argument(
        "--trans_prob",
        help="Transition probability of the HMM to change state from WT to CNV. Default is 0.001. Lower values prefer longer CNVs, higher values prefer shorter CNVs.",
        required=False,
        type=float,
        default=0.001,
    )
    parser_workflow_cnv_calling.add_argument(
        "--plot_regions",
        dest="plot_regions",
        help="If set, the CNV calling script plots heatmaps of the CNV called regions with the corresponding sample group taken as background.",
        required=False,
        action="store_true",
    )
    parser_workflow_cnv_calling.set_defaults(plot_regions=False)
    parser_workflow_cnv_calling.add_argument(
        "--cluster_configfile",
        help="Path to the cluster config file.",
        required=False,
        type=os.path.abspath,
        default=None,
    )

    parser_workflow_cnv_calling.add_argument(
        "--drmaa_mem",
        help="Number of megabytes used with drmaa. Suggested is 16000. If --drmaa_mem and --drmaa_time are given, this workflow will automatically use --drmaa in its snakemake call.",
        required=False,
        type=int,
        default=None,
    )

    parser_workflow_cnv_calling.add_argument(
        "--drmaa_time",
        help="Maximum number of hh:mm per job. Suggested is 4:00.",
        required=False,
        type=str,
        default=None,
    )

    parser_workflow_cnv_calling.add_argument(
        "--drmaa_jobs",
        help="Number of jobs. Default is 50.",
        required=False,
        type=int,
        default=50,
    )
    parser_workflow_cnv_calling.set_defaults(func=workflow_cnv_calling.workflow_cnv_calling)

    # =========================================================================
    #  workflow reassignment
    # =========================================================================
    parser_workflow_reassignment = subparsers.add_parser(
        "workflow_reassignment",
        description=(
            "Complete CNV-calling snakemake-workflow to run on data from a single sequencing panel (bed-file)."
        ),
    )
    parser_workflow_reassignment.add_argument(
        "-w",
        "--workdir",
        help="Path to the snakemake workdir.",
        required=True,
        type=os.path.abspath,
    )
    parser_workflow_reassignment.add_argument(
        "-r",
        "--reference",
        help="Path to the genomic reference.",
        required=True,
        type=os.path.abspath,
    )
    parser_workflow_reassignment.add_argument(
        "-m",
        "--metafile",
        help="Path to the file containing the meta information. It is a .tsv of the scheme [panel] [bams.txt] [bed.bed]. \
            It aligns each desired panel name with the corresponding .bam files and the .bed file.",
        required=True,
        type=os.path.abspath,
    )

    parser_workflow_reassignment.add_argument(
        "-c",
        "--coverages",
        help="Output file path to coverages matrix (e.g. coverages.tsv).",
        required=True,
        type=os.path.abspath,
    )

    parser_workflow_reassignment.add_argument(
        "-b",
        "--bedfile",
        required=True,
        type=os.path.abspath,
        help="Output file path to BED file that will contain the union of all given BED files",
    )

    parser_workflow_reassignment.add_argument(
        "--cores",
        help="Number of used cores in snakemake workflow. Default is 32.",
        required=False,
        type=int,
        default=32,
    )

    parser_workflow_reassignment.add_argument(
        "--cluster_configfile",
        help="Path to the cluster config file.",
        required=False,
        type=os.path.abspath,
        default=None,
    )

    parser_workflow_reassignment.add_argument(
        "--drmaa_mem",
        help="Number of megabytes used with drmaa. Suggested is 16000. If --drmaa_mem and --drmaa_time are given, this workflow will automatically use --drmaa in its snakemake call.",
        required=False,
        type=int,
        default=None,
    )

    parser_workflow_reassignment.add_argument(
        "--drmaa_time",
        help="Maximum number of hh:mm per job. Suggested is 4:00.",
        required=False,
        type=str,
        default=None,
    )
    parser_workflow_reassignment.add_argument(
        "--drmaa_jobs",
        help="Number of jobs. Default is 50.",
        required=False,
        type=int,
        default=50,
    )
    parser_workflow_reassignment.set_defaults(func=workflow_reassignment.workflow_reassignment)

    # =========================================================================
    #  visualize reassignment
    # =========================================================================
    parser_visualize_reassignmentment = subparsers.add_parser(
        "visualize_reassignment",
        description=("Interactive reassignment visualization"),
    )

    parser_visualize_reassignmentment.add_argument(
        "--host", default="0.0.0.0", help="Host to run server on, defaults to 0.0.0.0"
    )

    parser_visualize_reassignmentment.add_argument(
        "--port", type=int, default=8080, help="Port to run server on"
    )

    parser_visualize_reassignmentment.add_argument(
        "--debug", default=False, action="store_true", help="Whether or not to enable debugging"
    )

    parser_visualize_reassignmentment.add_argument(
        "--cache-dir", help="Optional path to cache directory, avoids repeating startup computation"
    )

    parser_visualize_reassignmentment.add_argument(
        "-m",
        "--metafile",
        help="Path to the file containing the meta information. It is a .tsv of the scheme [panel name] [path of bams.txt] [path of targets.bed]. \
            It aligns each desired panel name with the corresponding .bam files and the .bed file.",
        required=True,
        type=os.path.abspath,
    )

    parser_visualize_reassignmentment.add_argument(
        "-c",
        "--coverages",
        required=True,
        type=os.path.abspath,
        help="Matrix in tsv format containing coverage values. Is output of 'workflow_reassignment' step",
    )

    parser_visualize_reassignmentment.add_argument(
        "-b",
        "--bedfile",
        required=True,
        type=os.path.abspath,
        help="Path to BED file that will contain the union of all given BED files",
    )

    parser_visualize_reassignmentment.add_argument(
        "-d",
        "--new_panel_assignments_directory",
        help="Path to directory that will contain the output lists of paths to .bam-files, which were re-assigned to the given panels according to clustering.",
    )

    # parser_visualize_reassignmentment.add_argument(
    #    "-p",
    #    "--panels",
    #    required=True,
    #    type=str,
    #    help="Path to the directory that will contain the final lists of bam-files grouped by panels")

    # parser_visualize_reassignmentment.add_argument(
    #    "-a",
    #    "--batches",
    #    required=True,
    #    type=str,
    #    help="Path to the directory that will contain the final lists of bam-files grouped by batches",
    # )

    def viz_reassignment(args):
        """Helper function that launches the Dash webserver for reassignment visualization."""
        with tempfile.TemporaryDirectory() as tmpdir:
            from .viz_reassignment import settings

            if args.cache_dir:
                settings.CACHE_DIR = args.cache_dir
            else:
                settings.CACHE_DIR = os.path.join(tmpdir, "cache")
            os.makedirs(settings.CACHE_DIR, exist_ok=True)
            settings.setup_paths(args)

            settings.BATCH_OUTPUT_PATH = args.new_panel_assignments_directory

            from .viz_reassignment.app import app  # noqa
            from .viz_reassignment import store

            if settings.CACHE_PRELOAD_DATA:
                store.load_all_data(settings.reassign_SETTINGS)
            from .viz_reassignment import ui_plots

            ## XXX
            us = settings.reassign_SETTINGS
            data = store.load_all_data(us)
            ui_plots.plot_clustermap_clustering_as_base64(
                data, *store.compute_acluster(us), store.compute_clustercoldict(us)
            )
            # ui_plots.plot_clustermap_batches_as_base64(
            #    data, store.compute_acluster(us), store.compute_clustercoldict(us)
            # )
            ## XXX
            app.run_server(host=args.host, port=args.port, debug=args.debug)

    parser_visualize_reassignmentment.set_defaults(func=viz_reassignment)

    return parser


def main():
    if not misc.execs_available(REQUIRED_EXECUTABLES):
        logger.error("Missing some executables. The program will eventually fail.")
    else:
        logger.debug("External executables present: %s", ", ".join(REQUIRED_EXECUTABLES))
    parser = get_parser()
    args = parser.parse_args()
    if not hasattr(args, "func"):
        parser.print_help()
    else:
        logger.info("Starting analysis...")
        args.func(args)
        logger.info("All done. Have a nice day!")


if __name__ == "__main__":
    main()



import subprocess, tempfile, shlex
import pathlib
from logzero import logger


def create_config(configpath, args):
    with open(configpath, "wt") as f:
        print("workdir: '" + args.workdir + "'", file=f)
        print("panelname: '" + args.panelname + "'", file=f)
        print("bedfile: '" + args.bedfile + "'", file=f)
        print("bamsfile: '" + args.bamsfile + "'", file=f)
        print("reference: '" + args.reference + "'", file=f)
        print("kmer_align: '" + args.kmer_align + "'", file=f)
        # options
        print("expected_artefacts: '" + str(args.expected_artefacts) + "'", file=f)
        print("sample_score_factor: '" + str(args.sample_score_factor) + "'", file=f)
        print("minimum_group_sizes: '" + str(args.minimum_group_sizes) + "'", file=f)
        print("zscale: '" + str(args.zscale) + "'", file=f)
        print("size: '" + str(args.size) + "'", file=f)
        print("del_cutoff: '" + str(args.del_cutoff) + "'", file=f)
        print("dup_cutoff: '" + str(args.dup_cutoff) + "'", file=f)
        print("trans_prob: '" + str(args.trans_prob) + "'", file=f)
        print("plot_regions: '" + str(args.plot_regions) + "'", file=f)


def workflow_cnv_calling(args):
    configpath = tempfile.NamedTemporaryFile(suffix=".yml")
    # configpath = (
    #     pathlib.Path(__file__).absolute().parent
    #     / pathlib.Path("workflow")
    #     / pathlib.Path("config_cnv_calling.yml")
    # )
    # configpath.parent.mkdir(parents=True, exist_ok=True)
    create_config(configpath.name, args)
    arguments = [
        "snakemake",
        "--snakefile",
        str(
            pathlib.Path(__file__).absolute().parent
            / pathlib.Path("workflow")
            / pathlib.Path("Snakefile_cnv_calling")
        ),
        "--configfile",
        str(configpath.name),
    ]
    if args.cluster_configfile:
        arguments.append("--cluster-config")
        arguments.append(args.cluster_configfile)
    if args.drmaa_mem and args.drmaa_time and args.drmaa_jobs:
        arguments.append("-p")
        arguments.append("--drmaa")
        arguments.append(
            str(
                f'" --mem={args.drmaa_mem} --time={args.drmaa_time} --output={str(args.workdir)}/sge_log/%x-%j.log"'
            )
        )
        arguments.append("--jobs")
        arguments.append(str(args.drmaa_jobs))
    else:
        arguments.append("--cores")
        arguments.append(str(args.cores))
    # arguments.append("--rerun-incomplete")
    # arguments.append("--unlock")
    logger.info("snakemake arguments: %s" % (" ".join(arguments)))
    subprocess.check_call(arguments)

    # "panelname="+args.panelname,
    # "workdir="+args.workdir,
    # "bedfile="+args.bedfile,
    # "bamsfile="+args.bamsfile,
    # "reference="+args.reference,
    # "kmer_align="+args.kmer_align)



import base64
import io
import tempfile

import math
import matplotlib
import matplotlib.pyplot as plt
from matplotlib import cm
from matplotlib.colors import ListedColormap, LinearSegmentedColormap
from clearCNV import util

from logzero import logger

import seaborn as sns

sns.set_theme(color_codes=True)


from .cache import cache

# matplotlib.use('Agg')


@cache.memoize()
def plot_clustermap_panels_as_base64(data, panelcoldict):
    logger.info("Creating panels clustermap ...")
    n = min([data.X.shape[0], 1000])
    m = min([data.X.shape[1], 1000])
    plt.figure(figsize=(8, 5))
    sns.clustermap(
        data.X.iloc[:: math.ceil(data.X.shape[0] / n), :: math.ceil(data.X.shape[1] / m)],
        col_colors=list(
            map(
                lambda x: panelcoldict[x],
                data.samples["panel"].iloc[:: math.ceil(data.X.shape[1] / m)],
            )
        ),
        cbar_kws={"ticks": [0, 1]},
    )
    logger.info("... done creating panels clustermap.")
    with tempfile.NamedTemporaryFile(suffix=".png") as f:
        plt.savefig(f, format="png")
        f.flush()
        f.seek(0)
        return base64.b64encode(f.read()).decode("ascii")


@cache.memoize()
def plot_clustermap_clustering_as_base64(data, XD, cluster_panel_dict, clustercoldict):
    # assign_clustercolors_to_panels = ACP
    ACP = {i: data.panels.index(p) for i, p in enumerate(cluster_panel_dict.values())}
    logger.info("Creating clustering clustermap ...")
    n = min([data.X.shape[0], 1000])
    m = min([data.X.shape[1], 1000])
    plt.figure(figsize=(8, 5))
    sns.clustermap(
        data.X.iloc[:: math.ceil(data.X.shape[0] / n), :: math.ceil(data.X.shape[1] / m)],
        col_colors=list(
            map(
                lambda x: clustercoldict[ACP[x]],
                XD["clustering"].iloc[:: math.ceil(data.X.shape[1] / m)],
            )
        ),
        cbar_kws={"ticks": [0, 1]},
    )

    logger.info("... done creating clustering clustermap.")
    with tempfile.NamedTemporaryFile(suffix=".png") as f:
        plt.savefig(f, format="png")
        f.flush()
        f.seek(0)
        return base64.b64encode(f.read()).decode("ascii")


@cache.memoize()
def plot_clustermap_batches_as_base64(data, XD, clustercoldict):
    logger.info("Creating batches clustermap ...")
    n = min([data.X.shape[0], 1000])
    m = min([data.X.shape[1], 1000])
    plt.figure(figsize=(8, 5))
    D3 = util.normalize_within_exon(util.normalize_within_sample(data.D1))
    D4 = D3.applymap(lambda x: max([0, min([x, 2])]))
    sns.clustermap(
        D4.iloc[:: math.ceil(D4.shape[0] / n), :: math.ceil(D4.shape[1] / m)],
        col_colors=list(
            map(
                lambda x: clustercoldict[x],
                XD["clustering"].iloc[:: math.ceil(data.X.shape[1] / m)],
            )
        ),
        cbar_kws={"ticks": [0, 1]},
    )
    logger.info("... done creating batches clustermap.")
    with tempfile.NamedTemporaryFile(suffix=".png") as f:
        plt.savefig(f, format="png")
        f.flush()
        f.seek(0)
        return base64.b64encode(f.read()).decode("ascii")



import subprocess
import pathlib
import tempfile

from . import misc


def create_config(configpath, args):
    with open(configpath, "wt") as f:
        print("workdir: '" + str(args.workdir) + "'", file=f)
        print("metafile: '" + str(args.metafile) + "'", file=f)
        print("coverages: '" + str(args.coverages) + "'", file=f)
        print("cov_dir: '" + str(args.cov_dir) + "'", file=f)
        print("bedfile: '" + str(args.bedfile) + "'", file=f)
        print("bamsfile: '" + str(args.bamsfile) + "'", file=f)
        print("reference: '" + str(args.reference) + "'", file=f)


def workflow_reassignment(args):
    args.bamsfile = (pathlib.Path(args.workdir) / "allbams.txt").absolute()
    args.cov_dir = (pathlib.Path(args.workdir) / "covs").absolute()
    args.bedfile = pathlib.Path(args.bedfile).absolute()
    args.coverages = pathlib.Path(args.coverages).absolute()
    args.metafile = pathlib.Path(args.metafile).absolute()
    args.reference = pathlib.Path(args.reference).absolute()

    misc.prepare_reassignment(args)

    configpath = (
        pathlib.Path(__file__).absolute().parent
        / pathlib.Path("workflow")
        / pathlib.Path("config_reassignment.yml")
    )
    configpath.parent.mkdir(parents=True, exist_ok=True)
    create_config(configpath, args)
    arguments = [
        "snakemake",
        "--snakefile",
        str(
            pathlib.Path(__file__).absolute().parent
            / pathlib.Path("workflow")
            / pathlib.Path("Snakefile_reassignment")
        ),
        "--configfile",
        str(configpath),
        "--cores",
        str(args.cores),
    ]
    if args.cluster_configfile:
        arguments.append("--cluster-config")
        arguments.append(args.cluster_configfile)
    if args.drmaa_mem and args.drmaa_time and args.drmaa_jobs:
        arguments.append("-p")
        arguments.append("--drmaa")
        workdir_full_path = str(pathlib.Path(args.workdir).absolute())

        drmaa_string = f" --mem={args.drmaa_mem} --time={args.drmaa_time} --output={workdir_full_path}/sge_log/%x-%j.log "
        arguments.append(drmaa_string)
        arguments.append("-j")
        arguments.append(str(args.drmaa_jobs))
    # How to create the dir the best way?
    subprocess.check_call(["mkdir", "-p", f"{args.workdir}sge_log"])
    subprocess.check_call(["mkdir", "-p", f"{args.workdir}covs"])

    subprocess.check_call(arguments)



"""Handle access to the data."""

import typing

import os
import attr
import cattr
import pickle

import pathlib
import pandas as pd
import numpy as np
from math import sqrt

from collections import Counter
import subprocess

from sklearn.mixture import GaussianMixture
from sklearn.decomposition import PCA
from sklearn.manifold import TSNE
from sklearn.cluster import AgglomerativeClustering

import matplotlib.pyplot as plt
from matplotlib import cm
from matplotlib.colors import ListedColormap, LinearSegmentedColormap

from logzero import logger

from .. import util
from .. import cnv_arithmetics as ca
from .cache import cache
from . import settings
from . import ui_plots


#: Whether or not to pickle data (for development only).
PICKLE_DATA = False


@attr.s(auto_attribs=True, frozen=True)
class Data:
    BED: typing.Any
    D: typing.Any
    D1: typing.Any
    META: typing.Any
    X: typing.Any
    allsamples: typing.Any
    dropoutsamples: typing.Any
    samples: typing.Any
    panels: typing.Any
    bedsmatrix: typing.Any


def _load_coverage(path, sep="\t"):
    D0 = pd.read_csv(path, sep="\t", low_memory=False)
    if not (
        D0.columns[0] in ["chr", "Chr", "chromosome", "Chromosome"]
        and D0.columns[1] in ["start", "Start"]
        and D0.columns[2] in ["end", "End"]
    ):
        raise Exception(
            "File needs to be formatted like: chr\tstart\tend\tsample_0\tsample_1\tsample_n"
        )
    D0 = D0.iloc[:, 3:]
    return D0


# User must define number of batches
def _find_batches(xd, n=1):
    xds = xd[["X", "Y"]].to_numpy()
    xindex = xd.index
    GM = GaussianMixture(n_components=max([0, n]), n_init=10).fit(xds)
    # print(GM.bic(xds))
    xds_df = pd.DataFrame(xds, index=xd.index, columns=["X", "Y"])
    xds_df["batch"] = GM.predict(xds).astype(str)
    xds_df["panel"] = xd["panel"]
    # BIC = GM.bic(xds)
    return xds_df


@cache.memoize()
def load_all_data(us: settings.reassignSettings):
    if PICKLE_DATA and os.path.exists("all_data.bin"):
        with open("all_data.bin", "rb") as f:
            return pickle.load(f)
    logger.info("Starting to load data ...")
    logger.info("    - %s", settings.PATH_BEDFILE)
    BED = pd.read_csv(settings.PATH_BEDFILE, sep="\t", header=None).astype(str)
    BED = BED[[b.isdigit() for b in BED.iloc[:, 0]]]
    BED = BED.astype(int)

    logger.info("    - %s", settings.PATH_COVERAGES)
    D = _load_coverage(settings.PATH_COVERAGES)
    D = D.T[BED.index].T

    META = pd.read_csv(settings.PATH_METAFILE, sep="\t")
    META = META.sort_values(by="panel")

    logger.info("    - allsamples")
    allsamples = pd.DataFrame()
    for i in META.index:
        s = META.loc[i, :]
        bamfiles = [line.rstrip("\n") for line in open(s["bamsfiles"])]
        allsamples = pd.concat(
            [
                allsamples,
                pd.DataFrame(
                    list(
                        zip(
                            [s["panel"]] * len(bamfiles),
                            [pathlib.Path(bf).stem for bf in bamfiles],
                            [bf for bf in bamfiles],
                        )
                    )
                ),
            ]
        )
    allsamples = allsamples.set_index(1)
    allsamples.columns = ["panel", "path"]

    D = D[allsamples.index]

    d = D.columns[((D > 1).sum() == 0) | (D.sum() <= us.threshold)]
    dropoutsamples = pd.DataFrame(allsamples.T[d].T["path"])

    D1 = D.drop(columns=dropoutsamples.index)
    D.index = BED.apply(lambda row: "chr%s:%s" % (row[0], row[1]), axis=1)
    # X = D1 > us.threshold
    X = D1.ge((BED.iloc[:, 2] - BED.iloc[:, 1]) / us.threshold, axis=0)

    logger.info("    - bedsmatrix")
    samples = allsamples.drop(dropoutsamples.index)
    panels = sorted(set(META["panel"]))
    bedsmatrix = np.zeros((len(panels), BED.shape[0]))

    for j, panel in enumerate(panels):
        bed_panel_path = META.iloc[j, 2]
        # def merge_bedfiles(beds,bedfile):
        command = [
            "bedtools",
            "intersect",
            "-wa",
            "-a",
            settings.PATH_BEDFILE,
            "-b",
            bed_panel_path,
        ]
        p = subprocess.Popen(
            command, stdout=subprocess.PIPE, stderr=subprocess.PIPE, encoding="utf-8"
        )
        output, error = p.communicate()

        btargets = [ca.Target(*BED.iloc[i, :], "gene") for i in BED.index]
        ptargets = [ca.Target(*line.split("\t"), "gene") for line in output.split("\n")[:-1]]
        for i, t in enumerate(btargets):
            if t in ptargets:
                bedsmatrix[j][i] = 1
    bedsmatrix = pd.DataFrame(bedsmatrix, index=panels)

    logger.info("... done loading data.")

    result = Data(
        BED=BED,
        D=D,
        D1=D1,
        X=X,
        META=META,
        allsamples=allsamples,
        dropoutsamples=dropoutsamples,
        samples=samples,
        panels=panels,
        bedsmatrix=bedsmatrix,
    )
    if PICKLE_DATA:
        with open("all_data.bin", "wb") as f:
            pickle.dump(result, f)
    return result


@cache.memoize()
def compute_pca(us: settings.reassignSettings):
    logger.info("Computing PCA ...")
    data = load_all_data(us)
    pca_simple = PCA(n_components=2, random_state=us.pca_seed)
    pca2df = pd.DataFrame(pca_simple.fit_transform(data.X.dropna(0).T))
    pca2df.columns = ["X", "Y"]
    pca2df["panel"] = list(data.samples["panel"])
    logger.info("... done computing PCA.")
    return pca2df


@cache.memoize()
def compute_tsne(us: settings.reassignSettings):
    logger.info("Computing tSNE ...")
    data = load_all_data(us)
    pca = PCA(
        n_components=us.pca_components,
        random_state=us.pca_seed,
    )
    XT = pca.fit_transform(data.X.dropna(0).T)

    X_embedded = TSNE(n_components=2, random_state=us.pca_seed).fit_transform(XT)

    XD = pd.DataFrame(X_embedded)
    XD.columns = ["X", "Y"]
    XD["panel"] = list(data.samples["panel"])
    logger.info("... done computing tSNE.")
    return XD


@cache.memoize()
def compute_acluster(us: settings.reassignSettings):
    logger.info("Computing Agglomerative Clustering ...")
    data = load_all_data(us)
    XD = compute_tsne(us)
    clustering = AgglomerativeClustering(n_clusters=len(data.panels)).fit(XD[["X", "Y"]].to_numpy())
    # TODO: yikes, inplace update with cached data...
    # A: not really. 'XD["clustering"]' is a new column.
    XD["clustering"] = clustering.labels_
    # majority vote for panel assignments
    cluster_panel_dict = {
        i: Counter(data.samples[clustering.labels_ == i]["panel"]).most_common()[0][0]
        for i in range(len(data.panels))
    }
    XD["new_assignments"] = list(map(lambda x: cluster_panel_dict[x], clustering.labels_))
    logger.info("... done computing Agglomerative Clustering.")
    return (XD, cluster_panel_dict)


@cache.memoize()
def compute_panelcoldict(us: settings.reassignSettings):
    data = load_all_data(us)
    clustercolors = cm.get_cmap(us.colormap, len(data.panels))
    panelcoldict = {
        data.panels[i]: c for i, c in enumerate(clustercolors(np.linspace(0, 1, len(data.panels))))
    }
    return panelcoldict


@cache.memoize()
def compute_clustercoldict(us: settings.reassignSettings):
    data = load_all_data(us)
    clustercolors = cm.get_cmap(us.colormap, len(data.panels))
    # cluster ~ color ~ panel
    clustercoldict = {
        i: c for i, c in enumerate(clustercolors(np.linspace(0, 1, len(data.panels))))
    }
    return clustercoldict


@cache.memoize()
def compute_batches(us: settings.reassignSettings):
    data = load_all_data(us)
    XD, cluster_panel_dict = compute_acluster(us)
    XD.index = data.D1.columns
    batches = []
    for i, cluster in enumerate(sorted(set(XD["new_assignments"]))):
        x = data.D1.T[XD["new_assignments"] == cluster]
        x = x.fillna(0).T[x.median(axis=0) > us.threshold].T
        d = x[x.median(axis=1) <= us.threshold]
        x = x.T[x.index.difference(d.index)].T
        p = x.index
        x = pd.DataFrame(util.normalize(util.normalize(x, axis=1), axis=0)).T.fillna(1).T
        x.index = p

        # perform clustering
        pca = PCA(n_components=us.pca_components)
        xt = pca.fit_transform(x)

        x_embedded = TSNE(n_components=2).fit_transform(xt)

        xd = pd.DataFrame(x_embedded)
        xd.index = x.index
        xd.columns = ["X", "Y"]
        xd["panel"] = data.samples.loc[x.index, "panel"]

        batch_num_ = [int(val) for val in us.batch_num.split(",")]
        bf = batch_num_[i] if len(batch_num_) >= len(set(XD["new_assignments"])) else batch_num_[0]
        df = _find_batches(xd, bf)

        batches.append(df)
    return batches


@cache.memoize()
def save_results(us: settings.reassignSettings, n_clicks):
    data = load_all_data(us)
    XD, cluster_panel_dict = compute_acluster(us)
    # save new panel assignments
    # index changes after implicit filtering (?)
    XD.index = data.samples.index
    XD["paths"] = data.samples["path"]
    logger.info("saving new panel assignment ...")
    pathlib.Path(settings.BATCH_OUTPUT_PATH).absolute().mkdir(parents=True, exist_ok=True)
    for p in data.panels:
        bamspath = pathlib.Path(settings.BATCH_OUTPUT_PATH) / (str(p) + "_new_assigned.txt")
        XD[XD["new_assignments"] == p]["paths"].to_csv(
            bamspath, sep="\t", header=False, index=False
        )
    logger.info("... done saving new panel assignment.")
    batches = compute_batches(us)
    logger.info("saving new batch clusterings ...")
    for batch in batches:
        panel = set(batch["panel"]).pop()
        for i, x in enumerate(set(batch["batch"])):
            path = pathlib.Path(settings.BATCH_OUTPUT_PATH) / str(
                "%s_batch%.2d_%s.txt" % (panel, i, str(n_clicks))
            )
            logger.info("save batch file to: ", path)
            data.samples.T[batch[batch["batch"] == x].index].T["path"].to_csv(
                path, sep="\t", header=False, index=False
            )
    logger.info("... done saving new batch clusterings.")
    return True



"""Setup the SCelVis Dash application.

When importing this module, the app is built and configurd.  Thus, it is important that before
this module is imported, the values in ``.settings`` must already have been setup.
"""

import os

import dash
import flask
from flask import request, helpers
from logzero import logger


from . import callbacks, settings, store, cache
from ..__init__ import __version__
from .ui import build_layout

#: Path to assets.
ASSETS_FOLDER = os.path.join(os.path.dirname(__file__), "assets")

#: The Flask application to use.
app_flask = flask.Flask(__name__)

#: The Dash application to run.
app = dash.Dash(
    __name__,
    # Use our specific Flask app
    server=app_flask,
    # All files from "assets/" will be served as "/assets/*"
    assets_folder=ASSETS_FOLDER,
)

# Setup the cache.
cache.setup_cache(app)

# Set app title
app.title = "clear-CNV v%s" % __version__

# Serve assets locally
app.css.config.serve_locally = True
app.scripts.config.serve_locally = True

app.config.suppress_callback_exceptions = True

# Setup the application's main layout.
app.layout = build_layout()

# Register the callbacks with the app.
#
callbacks.register_control_to_buffer(app)
callbacks.register_buffer_to_graphs(app)
callbacks.register_save_button(app)



#!/usr/bin/env python
# coding: utf-8

import pandas as pd
import numpy as np
import multiprocessing as mp
import math
from clearCNV import util

# def matchscores(panel, intsv_path, matchscores_path, expected_cnv_rate, fast=True):
def matchscores(args):
    # parse args
    intsv_path = args.coverages
    matchscores_path = args.matchscores
    expected_cnv_rate = args.expected_artefacts
    CORES = args.cores if args.cores else mp.cpu_count()
    fast = True if args.fast else False

    D0 = util.load_dataframe(intsv_path)
    D3a = util.normalize_within_exon(
        util.normalize_within_sample(
            util.filter_by_coverage(
                D0[[not (s.startswith("X") or s.startswith("Y")) for s in D0.index]]
            )
        )
    )
    # only take every i-th target into account to speed up
    if fast:
        X = D3a.T[D3a.index[:: min([len(D3a.index), math.ceil(len(D3a.index) / 1000)])]].T
    else:
        X = D3a

    preMatchscores = pd.DataFrame(np.zeros((len(X.columns), len(X.columns))))

    def collect_result(result):
        nonlocal preMatchscores
        preMatchscores.iloc[:, result[0]] = result[1]

    pool = mp.Pool(CORES)

    print(
        f"calc matchscores with {CORES} threads at an expected artefact rate "
        f"of {expected_cnv_rate}"
    )
    for i in range(len(X.columns)):
        pool.apply_async(util.matchscore, args=(i, X, expected_cnv_rate), callback=collect_result)
    pool.close()
    pool.join()

    Matchscores = preMatchscores.add(preMatchscores.T)
    Matchscores.index, Matchscores.columns = X.columns, X.columns
    Matchscores.to_csv(matchscores_path, sep="\t", header=True, index=True)
    return Matchscores



"""Tests for the misc module.

If external programs are called then there are two variants of the test.  One has the suffix
"_integration" and the actual external program is called (must have been installed, of course).
The other has the suffix "_unit" and calls to external programs are mocked out and only the
mocking results are tested.
"""

from unittest.mock import MagicMock
from types import SimpleNamespace

from clearCNV import misc


def test_which():
    assert misc.execs_available(["ls", "sort", "bedtools", "bedops"])
    assert not misc.execs_available(["invented-and-bogus"])


def test_merge_bedfile_integration(tmp_path):
    path_out = str(tmp_path / "out.bed")
    args = SimpleNamespace(infile="tests/testdata/test_merge_bed_input.bed", outfile=path_out)
    misc.merge_bedfile(args)
    with open("tests/testdata/test_merge_bed_output.bed", "rt") as f:
        expected = f.read()
    with open(path_out, "rt") as f:
        actual = f.read()
    assert expected == actual


def test_merge_bedfile_unit(mocker, tmp_path):
    mock_popen = MagicMock()
    mocker.patch("clearCNV.misc.subprocess.Popen", mock_popen)
    args = SimpleNamespace(
        infile="tests/testdata/test_merge_bed_input.bed", outfile=str(tmp_path / "out.bed")
    )
    misc.merge_bedfile(args)
    assert mock_popen.call_count == 2


def test_merge_bedfiles_integration(tmp_path):
    path_out = str(tmp_path / "out.bed")
    misc.merge_bedfiles(
        [
            "tests/testdata/test_merge_beds_input_0.bed",
            "tests/testdata/test_merge_beds_input_1.bed",
        ],
        path_out,
    )
    with open("tests/testdata/test_merge_beds_output.bed", "rt") as f:
        expected = f.read()
    with open(path_out, "rt") as f:
        actual = f.read()
    assert expected == actual


def test_merge_bedfiles_unit(mocker, tmp_path):
    mock_popen = MagicMock()
    mocker.patch("clearCNV.misc.subprocess.Popen", mock_popen)
    misc.merge_bedfiles(
        [
            "tests/testdata/test_merge_beds_input_0.bed",
            "tests/testdata/test_merge_beds_input_1.bed",
        ],
        str(tmp_path / "out.bed"),
    )
    assert mock_popen.call_count == 3



#!/usr/bin/env python

import contextlib
import pathlib
import subprocess
import tempfile
import typing

from logzero import logger
import pandas as pd


def execs_available(execs: typing.List[str]) -> None:
    ok = True
    for exec in execs:
        res = subprocess.run(["which", exec], stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        if res.returncode != 0:
            logger.warn("Could not find executable %s via `which %s`", exec, exec)
            ok = False
    return ok


def merge_bedfile(args):
    # TODO: used anywhere besides tests?
    # Yes! Each time target counts are created from bam files!
    cmd_merge = ["bedtools", "merge", "-i", args.infile, "-c", "4", "-o", "collapse"]
    cmd_sort = ["sort", "-V", "-k1,1", "-k2,2"]
    with open(args.outfile, "wt") as f:
        with subprocess.Popen(cmd_merge, stdout=subprocess.PIPE, shell=False) as proc_merge:
            with subprocess.Popen(
                cmd_sort, stdin=proc_merge.stdout, stdout=f, shell=False
            ) as proc_sort:
                proc_merge.wait()
                proc_sort.wait()


def _popen(cmd, stdin=None, stdout=None):
    """Simple wrapper for Popen to shorten things and in particular disables using the shell."""
    stdout = stdout or subprocess.PIPE
    return subprocess.Popen(cmd, stdin=stdin, stdout=stdout, shell=False)


# =============================================================================
#  reassignment - preparations


def merge_bedfiles(beds, bedfile):
    pathlib.Path().absolute().parent.mkdir(parents=True, exist_ok=True)

    cmd_merge = ["bedops", "--merge", *map(str.rstrip, beds)]
    cmd_sort = ["sort", "-V", "-k1,1", "-k2,2"]
    cmd_merge2 = ["bedtools", "merge"]

    with contextlib.ExitStack() as stack:
        # Open output file and start process pipeline.
        f = stack.enter_context(open(bedfile, "wt"))
        p_merge = _popen(cmd_merge)
        p_sort = _popen(cmd_sort, p_merge.stdout)
        p_merge2 = _popen(cmd_merge2, p_sort.stdout, f)
        # Wait for completion.
        p_merge.wait()
        p_sort.wait()
        p_merge2.wait()


def prepare_reassignment(args):
    META = pd.read_csv(args.metafile, sep="\t")
    # prepare bamsfile
    pathlib.Path(args.bamsfile).absolute().parent.mkdir(parents=True, exist_ok=True)
    f = open(args.bamsfile, "wt")
    L = [[l.rstrip("\n") for l in open(bam)] for bam in META.iloc[:, 1]]
    for b in set([val for l in L for val in l]):
        print(b, file=f)
    f.close()
    merge_bedfiles(list(META["bedfiles"]), args.bedfile)


# =============================================================================
#  ANNOTATIONS


def __gc_content(reference, bedfile, GC_annotation):
    # with path_out.open("wt") as f:
    with open(GC_annotation, "wt") as f:
        subprocess.check_call(["bedtools", "nuc", "-fi", reference, "-bed", bedfile], stdout=f)


def __mappability(bedfile, kmer_align, mappability_annotation):
    with open(mappability_annotation, "wt") as f:
        subprocess.check_call(["bedtools", "coverage", "-a", bedfile, "-b", kmer_align], stdout=f)


def __merge_annotations(mappability_annotation, GC_annotation, annotations):
    pd.read_csv(mappability_annotation, sep="\t", header=None).iloc[:, [0, 1, 2, 3, 6, 7]].join(
        pd.read_csv(GC_annotation, sep="\t").iloc[:, 5]
    ).to_csv(annotations, sep="\t", header=None, index=False)


def create_annotations_file(args):
    with tempfile.TemporaryDirectory() as tmp_dir:
        d = pathlib.Path(tmp_dir)
        GC_annotation = d / pathlib.Path("GC_annotation.bed")
        mappability_annotation = d / pathlib.Path("mappability_annotation.bed")
        __gc_content(args.reference, args.bedfile, GC_annotation)
        __mappability(args.bedfile, args.kmer_align, mappability_annotation)
        __merge_annotations(mappability_annotation, GC_annotation, args.annotations)


# =============================================================================
#  COVERAGES


def count_pair(args):
    with open(args.rtbed, "wt") as f:
        subprocess.call(
            ["bedtools", "multicov", "-p", "-q", "20", "-bams", args.inbam, "-bed", args.bedfile],
            stdout=f,
        )


def merge_rtbeds(args):
    D = pd.read_csv(args.bedfile, sep="\t", header=None)
    dcols = D.shape[1]
    for i in range(len(args.rtbeds)):
        D[dcols + i] = pd.read_csv(args.rtbeds[i], sep="\t", header=None)[dcols]
    D.columns = ["chr", "start", "end", "gene"][:dcols] + [
        pathlib.Path(s).stem for s in args.rtbeds
    ]
    D.to_csv(args.coverages, sep="\t", index=False)



#!/usr/bin/env python
# coding: utf-8


import pathlib
import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
import math
from collections import OrderedDict
from sklearn.cluster import KMeans

import plotly.graph_objects as go
import plotly
from plotly.subplots import make_subplots

# def visualize(analysis_dir, ratio_scores_path, z_scores_path, annotated_path, htmp_size):
def visualize(args):
    # argparsing
    analysis_dir = args.analysis_directory
    ratio_scores_path = args.ratio_scores
    z_scores_path = args.z_scores
    annotated_path = args.annotated
    htmp_size = args.size

    # load data
    annotated = pd.read_csv(annotated_path, sep="\t", low_memory=False, header=None)
    annotated.columns = ["chr", "start", "end", "gene", "size", "mapp", "gc"]
    annotated.index = [
        str(annotated["chr"][i])
        + "_"
        + str(annotated["start"][i])
        + "_"
        + str(annotated["end"][i])
        + "_"
        + str(annotated["gene"][i])
        for i in range(len(annotated.index))
    ]
    ratio_scores_df = pd.read_csv(ratio_scores_path, sep="\t", low_memory=False, index_col=0)
    z_scores_scaled_df = pd.read_csv(z_scores_path, sep="\t", low_memory=False, index_col=0)
    RS = ratio_scores_df.fillna(1.0).copy()

    def splits(index, htmp_size):
        indexchrs = [s.split("_")[0] for s in index]
        chrs = list(OrderedDict.fromkeys(indexchrs))
        chrstarts = [indexchrs.index(s) for s in chrs]
        chrstarts.append(len(indexchrs))
        arr = np.abs(np.array(chrstarts) - (len(index) / 2))
        split_pos = list(arr).index(min(arr)) + 1

        if len(chrs) <= 1 or len(index) <= htmp_size:
            return [index]
        else:
            return [
                *splits(index[: chrstarts[:split_pos][-1]], htmp_size),
                *splits(index[chrstarts[:split_pos][-1] :], htmp_size),
            ]

    def get_plots(df, annotated):
        n_clusters = math.ceil(len(df.columns) / 40)
        kmeans = KMeans(n_clusters=n_clusters, random_state=0).fit(
            df[df.std(axis=1) < np.nanquantile(df.std(axis=1), 0.95)].T
        )

        clustered_samples = pd.DataFrame(list(zip(df.columns, kmeans.labels_))).sort_values(by=1)[0]

        df = df[clustered_samples].copy()

        htmp = go.Heatmap(
            z=[np.clip(df.iloc[i, :], 0, 2) for i in range(len(df.index))],
            x=df.columns,
            y=df.index,
            hoverongaps=False,
        )

        mgc = go.Scatter(y=df.index, x=annotated.loc[df.index, "gc"])
        mmapp = go.Scatter(y=df.index, x=annotated.loc[df.index, "mapp"])
        msize = go.Scatter(y=df.index, x=annotated.loc[df.index, "size"])

        sp = make_subplots(
            rows=1,
            cols=4,
            subplot_titles=("ratios", "mapp", "gc", "log_size"),
            column_widths=[0.85, 0.05, 0.05, 0.05],
            shared_yaxes=True,
        )

        sp.add_trace(htmp, row=1, col=1)
        sp.add_trace(mmapp, row=1, col=2)
        sp.add_trace(mgc, row=1, col=3)
        sp.add_trace(msize, row=1, col=4)

        sp.update_layout(
            width=len(df.columns) * 3 + 300,
            height=len(df.index) * 3 + 300,
        )
        return sp

    print("plotting heatmaps...")
    # parallelize?
    for i, split in enumerate(splits(RS.index, htmp_size)):
        plotly.offline.plot(
            get_plots(RS.loc[split, :], annotated),
            filename=str(pathlib.Path(analysis_dir) / f"ratio_scores_extended_{i}.html"),
        )
        print(f"ratio_scores_extended_{i}.html finished")

    # analyze relation GC content and st. dev. of z-scores after grouping
    print("analyze relation GC content and st. dev. of z-scores after grouping.")

    z_scores_scaled_df["std"] = [
        z_scores_scaled_df.loc[i, :].std() for i in z_scores_scaled_df.index
    ]
    annotated["std"] = z_scores_scaled_df["std"]
    z_scores_scaled_df = z_scores_scaled_df.drop(columns="std")

    ratio_scores_df["std"] = [ratio_scores_df.loc[i, :].std() for i in ratio_scores_df.index]
    annotated["ratio"] = ratio_scores_df["std"]
    ratio_scores_df = ratio_scores_df.drop(columns="std")

    plt.figure(figsize=(8, 5))
    plt.scatter(annotated["std"], annotated["gc"], marker=".", alpha=0.1, label="sd of z-scores")
    plt.ylabel("GC content")
    plt.xlabel("sd of z-scores")
    plt.title("GC content vs z-scores after grouping")
    plt.legend()
    plt.savefig(pathlib.Path(analysis_dir) / "ANALYSIS_GC_vs_SD.pdf", format="pdf")

    plt.figure(figsize=(8, 5))
    plt.scatter(
        annotated["ratio"], annotated["gc"], marker=".", alpha=0.1, label="sd of ratio scores"
    )
    plt.ylabel("GC content")
    plt.xlabel("sd of ratios")
    plt.title("GC content vs sd of ratio scores after grouping")
    plt.legend()
    plt.savefig(str(pathlib.Path(analysis_dir) / "ANALYSIS_ratio_vs_gc.pdf"), format="pdf")



"""Setup of the Flask cache."""

from flask_caching import Cache
from logzero import logger

from . import settings

#: The cache configuration, built from ``.settings``.
CACHE_CONFIG = {
    "DEBUG": settings.DEBUG,
    "CACHE_TYPE": settings.CACHE_TYPE,
    "CACHE_DEFAULT_TIMEOUT": settings.CACHE_DEFAULT_TIMEOUT,
    "CACHE_DIR": settings.CACHE_DIR,
    "CACHE_REDIS_URL": settings.CACHE_REDIS_URL,
}

#: The global cache instance.
cache = Cache()


def setup_cache(app):
    """Setup the Dash app's Flask app with the cache."""
    logger.info("Using cache configuration %s", CACHE_CONFIG)
    cache.init_app(app.server, config=CACHE_CONFIG)



"""Registration of callbacks for the UI."""

import base64
import os.path
import uuid
import json

import dash
import dash_html_components as html
from dash.dependencies import Input, Output, State
import pandas as pd
from logzero import logger
from werkzeug.utils import secure_filename

from . import ui, settings, store
from .settings import reassignSettings, BATCH_OUTPUT_PATH


def register_control_to_buffer(app):
    """Register callsbacks for controls."""

    @app.callback(
        Output("buffer-cluster-params", "children"),
        [
            Input("input-thresh", "value"),
            Input("input-pca-components", "value"),
            Input("input-batch-factor", "value"),
        ],
    )
    def parameters_to_json(threshold, pca_components, batch_num):
        return json.dumps(
            {
                "threshold": threshold,
                "pca_components": pca_components,
                "batch_num": batch_num,
            }
        )


def register_buffer_to_graphs(app):
    """Register callsbacks for controls."""

    @app.callback(
        [
            Output("graph-pca", "figure"),
            Output("graph-tsne", "figure"),
            Output("graph-agg-clust", "figure"),
            Output("container-image-cluster-panels", "children"),
            Output("container-image-cluster-clustering", "children"),
            # Output("container-image-batch-clustering", "children"),
            Output("container-image-batch-separation", "children"),
        ],
        [Input("buffer-cluster-params", "children")],
    )
    def parameters_to_json(params_str):
        us = reassignSettings(**json.loads(params_str))
        return [
            ui.render_pca(us),
            ui.render_tsne(us),
            ui.render_agg_clust(us),
            ui.render_image_cluster_panels(us),
            ui.render_image_cluster_clustering(us),
            # ui.render_image_batches_clustering(us),
            ui.render_images_batch_separation(us),
        ]


def register_save_button(app):
    """Register button for the save button."""

    @app.callback(
        Output("buffer-save-batches-text", "children"),
        [
            Input("input-save-batches-button", "n_clicks"),
            Input("buffer-cluster-params", "children"),
        ],
    )
    def on_click_button(n_clicks, params_str):
        if n_clicks is not None:
            us = reassignSettings(**json.loads(params_str))
            print("clicked save", n_clicks, us)
            # how to get us ????
            store.save_results(us, n_clicks)
            return "%d: saving to %s" % (n_clicks or 0, BATCH_OUTPUT_PATH)



import os.path
import setuptools

import versioneer


def parse_requirements(path):
    """Parse ``requirements.txt`` at ``path``."""
    requirements = []
    with open(path, "rt") as reqs_f:
        for line in reqs_f:
            line = line.strip()
            if line.startswith("-r"):
                fname = line.split()[1]
                inner_path = os.path.join(os.path.dirname(path), fname)
                requirements += parse_requirements(inner_path)
            elif line != "" and not line.startswith("#"):
                requirements.append(line)
    return requirements


with open("README.md") as readme_file:
    readme = readme_file.read()

with open("HISTORY.md") as history_file:
    history = history_file.read()


install_requirements = parse_requirements("requirements/base.txt")


setuptools.setup(
    name="clearCNV",
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    author="Vinzenz May",
    author_email="Vinzenz.May@mdc-berlin.de",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Topic :: Scientific/Engineering :: Bio-Informatics",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
    ],
    description="Clinical sequencing panel CNV caller and visualizer",
    long_description=readme + "\n\n" + history,
    long_description_content_type="text/markdown",
    include_package_data=True,
    url="https://github.com/bihealth/clear-CNV",
    packages=setuptools.find_packages(),
    entry_points={
        "console_scripts": [
            "clearCNV = clearCNV.__main__:main",
        ],
    },
    install_requires=install_requirements,
    python_requires=">=3.7",
    zip_safe=False,
)



#!/usr/bin/env python
# coding: utf-8

# In[2]:


import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
import math
import re
from sklearn.neighbors import KernelDensity
from sklearn.linear_model import LinearRegression
import seaborn as sns
import scipy.cluster.hierarchy as hc
from hmmlearn import hmm
from clearCNV import cnv_arithmetics as ca


def load_dataframe(path, sep="\t", annotated=None):
    D0 = pd.read_csv(path, sep="\t", low_memory=False)
    D0["index"] = [
        str(D0.iloc[i, 0])
        + "_"
        + str(D0.iloc[i, 1])
        + "_"
        + str(D0.iloc[i, 2])
        + "_"
        + str(D0.iloc[i, 3])
        for i in range(len(D0.index))
    ]
    if not (
        D0.columns[0] in ["chr", "Chr", "chromosome", "Chromosome"]
        and D0.columns[1] in ["start", "Start"]
        and D0.columns[2] in ["end", "End"]
        and D0.columns[3] in ["gene", "Gene"]
    ):
        raise Exception(
            "File needs to be formatted like: chr\tstart\tend\tgene\tsample_0\tsample_1\tsample_n"
        )
    D0.set_index("index", inplace=True)
    D0 = D0.iloc[:, 4:]
    D0.index = [s[3:] if s.startswith("chr") else s for s in D0.index]
    if annotated:
        D0 = D0.mul(annotated["mapp"] >= 1.0, axis=0)
    return D0


def load_annotated(path, header=["chr", "start", "end", "gene", "size", "mapp", "gc"]):
    annotated = pd.read_csv(path, sep="\t", low_memory=False, header=None)
    annotated.columns = header
    annotated.index = [
        str(annotated["chr"][i])
        + "_"
        + str(annotated["start"][i])
        + "_"
        + str(annotated["end"][i])
        + "_"
        + str(annotated["gene"][i])
        for i in range(len(annotated.index))
    ]
    annotated.index = [s[3:] if s.startswith("chr") else s for s in annotated.index]
    return annotated


def trimmed_std(df, axis=0, alpha=0.05):
    if axis not in (0, 1):
        raise ValueError("Invalid axis %d" % axis)
    if axis == 0:
        n = len(df.index)
    else:
        n = len(df.columns)
    df = df.copy()
    df.values.sort(axis=axis)
    lower = math.ceil(n * alpha)
    upper = math.floor(n * (1.0 - alpha))
    if axis == 0:
        return np.asarray(np.std(df.iloc[lower:upper, :], axis=axis, ddof=1))
    else:
        return np.asarray(np.std(df.iloc[:, lower:upper], axis=axis, ddof=1))


def trimmed_std_np(np_matrix, axis=1, alpha=0.1):
    D = np.sort(np_matrix, axis=axis)
    if axis:
        S = D[:, math.ceil(D.shape[axis] * alpha / 2) : -math.ceil(D.shape[axis] * alpha / 2)]
    else:
        S = D[math.ceil(D.shape[axis] * alpha / 2) : -math.ceil(D.shape[axis] * alpha / 2), :]
    return np.std(np.asarray(S), ddof=1, axis=axis)


def filter_by_coverage(D0_, threshold=5):
    # samples
    D0_ = D0_.transpose()[D0_.median(axis=0) > threshold].transpose()
    # exons
    D0_ = D0_[D0_.median(axis=1) > threshold]
    return D0_


def trim(D, axis=0, alpha=0.1):
    if axis:
        return D[
            trimmed_std_np(D, axis=axis) < np.quantile(trimmed_std_np(D, axis=axis), 1 - alpha), :
        ]
    else:
        return D[
            :, trimmed_std_np(D, axis=axis) < np.quantile(trimmed_std_np(D, axis=axis), 1 - alpha)
        ]


# within sample
def normalize(D, axis=1):
    return np.apply_along_axis(lambda v: v / np.median(v[np.nonzero(v)]), axis, np.asarray(D))


def normalize_within_exon(D0_):
    c, i = D0_.columns, D0_.index
    x = np.apply_along_axis(lambda v: v / np.median(v[np.nonzero(v)]), 1, np.asarray(D0_))
    return pd.DataFrame(x, columns=c, index=i)


def normalize_within_sample(D0_):
    c, i = D0_.columns, D0_.index
    x = np.apply_along_axis(lambda v: v / np.median(v[np.nonzero(v)]), 0, np.asarray(D0_))
    return pd.DataFrame(x, columns=c, index=i)


def normalize_chromosomewise(DA, INDEX):
    chroms = np.array([s.split("_")[0] for s in INDEX[DA[1]]])
    chroms_set = sorted([int(s) for s in set(chroms)])
    DD = np.zeros(DA[0].shape)
    chroms_mat = np.zeros((len(chroms), len(set(chroms)))).astype(bool)
    for i, c in enumerate(chroms_set):
        chroms_mat[:, i] = [v == str(c) for v in chroms]
    for i, c in enumerate(chroms_set):
        x = normalize(DA[0][chroms_mat[:, i]], axis=1)
        DD[chroms_mat[:, i], :] = x + 1 - np.median(x, axis=0)
    return DD


def matchscore(i, df, expected_cnv_rate):
    scores = np.zeros(len(df.columns))
    for j in range(len(df.columns)):
        if i > j:
            scores[j] = np.sum(
                sorted(abs(df.iloc[:, i] - df.iloc[:, j]))[
                    0 : math.ceil((1 - expected_cnv_rate) * len(df.index))
                ]
            ) / ((1 - expected_cnv_rate) * len(df.index))
    return (i, scores)


def print_clustermap(df, path, title, fileformat="pdf"):
    linkage = hc.linkage(df, method="average")
    clumap = sns.clustermap(df, row_linkage=linkage, col_linkage=linkage, figsize=(12, 12))
    ax = clumap.ax_heatmap
    ax.set_yticks(ticks=[])
    ax.set_xticks(ticks=[])
    plt.title(title)
    plt.savefig(path, format=fileformat)


# def hmm_predict(D):
#    df = D.copy()
#    model = hmm.GaussianHMM(n_components=3, covariance_type="full")
#    model.startprob_ = np.array([0.0001, 0.9998, 0.0001])
#    model.transmat_ = np.array(
#        [[0.9999, 0.0001, 0.0], [0.0001, 0.9998, 0.0001], [0.0, 0.0001, 0.9999]]
#    )
#    model.means_ = np.array([[-2], [0.0], [2]])
#    model.covars_ = np.tile(np.identity(1), (3, 1, 1))
#    for col in D.columns:
#        df[col] = model.predict([[s] for s in D[col]])
#    return df


def escore(target, rs):
    k = math.ceil(len(target) * 0.025)
    a = np.array(sorted(target)[k:-k]).reshape(-1, 1)
    kde = KernelDensity(kernel="gaussian", bandwidth=0.5).fit(a)
    s = np.linspace(0, 2, 100)
    e = kde.score_samples(s.reshape(-1, 1))
    n = np.where(e == max(e))[0][0]
    # right side
    if rs >= 1:
        x = s[n:].reshape((-1, 1))
        y = e[n:]
        Lmodel = LinearRegression().fit(x, y)
        return (np.exp(-Lmodel.predict(np.array(rs).reshape(-1, 1))) - 1)[0]
    # left side
    else:
        x = s[:n].reshape((-1, 1))
        y = e[:n]
        Lmodel = LinearRegression().fit(x, y)
        return (-np.exp(-Lmodel.predict(np.array(rs).reshape(-1, 1))) + 1)[0]


# takes np.array as input objects
def escore_sample(D, sample):
    rs = np.zeros(len(sample))
    for i, t in enumerate(sample):
        rs[i] = escore(D[:, i], t)
    return rs


def pcolor(D, factor=0.08):
    plt.figure(figsize=(factor * len(D.columns), factor * len(D.index)))
    plt.pcolor(D)
    plt.show()


def npcolor(D, factor=0.08):
    pcolor(pd.DataFrame(D), factor)


def hmm_scores(
    X,
    means=np.array([[-3.5], [0.0], [4.5]]),
    transitionprobs=np.array([[0.999, 0.001, 0.0], [0.001, 0.998, 0.001], [0.0, 0.001, 0.999]]),
):
    model = hmm.GaussianHMM(n_components=3, covariance_type="full")
    model.startprob_ = np.array([0.001, 0.998, 0.001])
    model.transmat_ = np.array(transitionprobs)
    model.means_ = np.array(means)
    model.covars_ = np.tile(np.identity(1), (3, 1, 1))
    return model.predict([np.array([s]) for s in X])


def score_cnv(start, end, z_scores):
    x = np.abs(z_scores[start:end])
    return np.mean(x)  # sum(np.log(x)/np.log(sum(x)))


def rscore_cnv(start, end, r_scores):
    x = r_scores[start:end]
    return np.mean(x)


# hmm_preds,z_scores,r_scores,index = HMM[9],z_scores_scaled[9],ratio_scores[9],INDEX
def merge_score_cnvs(hmm_preds, z_scores, r_scores, index, del_cutoff, dup_cutoff):
    CNVs = []
    length = 0 if hmm_preds[0] == 1 else 1
    start = 0
    for i in range(1, len(hmm_preds)):
        if hmm_preds[i] != hmm_preds[i - 1]:
            # w -> cnv
            if hmm_preds[i - 1] == 1:
                length = 1
                start = i
            # cnv -> w or cnv
            if hmm_preds[i - 1] != 1:
                rs = rscore_cnv(start, i, r_scores)
                # ensure that the ratio score is sufficiently high
                if rs <= del_cutoff or rs >= dup_cutoff:
                    CNVs.append(
                        ca.CNV(
                            *index[start].split("_")[:2],
                            *index[i - 1].split("_")[2:4],
                            "DUP" if hmm_preds[i - 1] == 2 else "DEL",
                            length - 1,
                            score_cnv(start, i, z_scores),
                            rs
                        )
                    )
                # cnv -> cnv
                if abs(hmm_preds[i] - hmm_preds[i - 1]) > 1:
                    length = 1
                    start = i
        if hmm_preds[i] != 1:
            length += 1
    return CNVs


def center_samples(D, limit=50):
    INDEX, SAMPLES = D.index, D.columns
    if D.shape[0] > limit:
        D = normalize(D, axis=0)
    D = pd.DataFrame(normalize(D, axis=1) + 1 - np.median(normalize(D, axis=1), axis=0))
    D.index, D.columns = INDEX, SAMPLES
    return D


def center_samples_np(D, limit=50):
    if D[0].shape[0] > limit:
        D = normalize(D, axis=0)
    return normalize(D, axis=1) + 1 - np.median(normalize(D, axis=1), axis=0)


def calling_cnv(index_sample, D, MS, index, EXPECTED_CNV_RATE, SENSITIVITY=0.7):
    sample_group = MS.iloc[index_sample]
    # per target median normalized sample
    # select sample group
    D1 = D[:, sample_group]
    # normalize sample per target
    ratio_values = D[:, index_sample] / np.median(D1, axis=1)
    # normalize sample group per target
    D2 = normalize(D1, axis=1)
    np.nan_to_num(D2, copy=False, nan=1.0)
    D3 = trim(D2, axis=0, alpha=0.1)

    # new metric instead of z-scores
    zscores = (ratio_values - 1.0) / (
        trimmed_std_np(D3, axis=1, alpha=EXPECTED_CNV_RATE) ** (2 - SENSITIVITY)
    )

    sample_score = 1.0 / (
        trimmed_std_np(zscores[:, None], axis=0, alpha=EXPECTED_CNV_RATE)[0] ** (2 - SENSITIVITY)
    )
    # print(index_sample,D1.shape,sample_score)
    z_scores_scaled = zscores * sample_score
    return (index_sample, index, sample_score, ratio_values, z_scores_scaled)


def turntransform(v):
    v -= min(v)
    v /= max(v)
    v += np.arange(1, 0, -1 / len(v))
    return v


def select_region(region, df, sample, matchscores_bools_selected, buffer=2):
    cols = df.columns
    if sample:
        # cols = np.append(np.array(df.loc[:,matchscores_bools_selected.loc[:,sample]].columns),sample)
        cols = np.array(df.loc[:, matchscores_bools_selected.loc[:, sample]].columns)
    target = re.split("[^0-9]", region.replace(" ", ""))
    index = np.array(
        [
            (
                c[0] == target[0]
                and int(c[1]) >= int(target[1]) - 5
                and int(c[2]) <= int(target[2]) + 5
            )
            for c in [s.split("_") for s in df.index]
        ]
    )
    n = index.shape[0]
    lower = np.argmax(index == 1)
    upper = np.argmax(index[lower:] == 0) + lower
    index[max(0, lower - buffer) : min(upper + buffer, n)] = True
    # df.loc[index,cols]
    r = df.loc[index, cols].copy()
    if not sample:
        return r
    spacer = pd.DataFrame(
        [float(round(np.mean(r.mean())))] * r.shape[0], index=r.index, columns=["spacer"]
    )
    rsample = pd.DataFrame(df.loc[index, sample])
    return pd.concat([r.T, spacer.T, rsample.T]).T



"""Settings for the Dash-based visual reassignment.

While it is general best practice not to have global state such as this settings module, the
Flask-based architecture of Dash forces us to use it.
"""

import pathlib

import attr
import cattr

#: Text to display on top left.
HOME_BRAND = "clear-CNV"

#: Base input path.
# PATH_BASE = None

#: Path to metafile.
PATH_METAFILE = None

#: Path to coverages.
PATH_COVERAGES = None

#: Path to targets BED.
PATH_BEDFILE = None

#: Path to batch definition.
BATCH_OUTPUT_PATH = None

#: Helper function to setup the paths based on an input path.
def setup_paths(args):
    global PATH_METAFILE, PATH_COVERAGES, PATH_BEDFILE, BATCH_OUTPUT_PATH
    # PATH_BASE = pathlib.Path(path)
    PATH_METAFILE = args.metafile
    PATH_COVERAGES = args.coverages
    PATH_BEDFILE = args.bedfile
    BATCH_OUTPUT_PATH = args.new_panel_assignments_directory


#: Whether or not to preload data.
CACHE_PRELOAD_DATA = True


@attr.s(frozen=True, auto_attribs=True)
class reassignSettings:
    threshold: int = 50
    pca_components: int = 20
    batch_num: str = "2"
    pca_seed: int = 100
    colormap = "tab20"


#: Configuration for reassignment.
reassign_SETTINGS = reassignSettings()

#: Whether or not to enable debugging.
DEBUG = False

#: The type of the cache to use from {"filesystem", "redis"}
CACHE_TYPE = "filesystem"
#: Default cache timeout.
CACHE_DEFAULT_TIMEOUT = 300
#: For "filesystem" cache: directory to store the cache in.
CACHE_DIR = None
#: For "redis" cache: the URL to use for connecting to the cache.
CACHE_REDIS_URL = None
#: whether or not to preload data on startup
CACHE_PRELOAD_DATA = True

#: Output path for batch separation.
# BATCH_OUTPUT_PATH = None



#!/usr/bin/env python
# coding: utf-8

import pathlib

import pandas as pd
import numpy as np

import multiprocessing as mp
import matplotlib.pyplot as plt
from sklearn.cluster import KMeans

from clearCNV import util
from clearCNV import cnv_arithmetics as ca


def cnv_calling(args):
    # argparsing
    intsv_path = args.coverages
    analysis_dir = args.analysis_directory
    matchscores_path = args.matchscores
    calls_path = args.cnv_calls
    ratio_scores_path = args.ratio_scores
    z_scores_path = args.z_scores
    EXPECTED_CNV_RATE = args.expected_artefacts
    SAMPLE_SCORE_FACTOR = args.sample_score_factor
    MINIMUM_SAMPLE_GROUP = args.minimum_group_sizes
    ZSCALE = args.zscale if args.zscale >= 0 and args.zscale <= 2 else 0.65
    CORES = min([args.cores, mp.cpu_count()]) if args.cores else mp.cpu_count()
    DUP_CUTOFF = args.dup_cutoff  # 1.35
    DEL_CUTOFF = args.del_cutoff  # 0.75
    TRANSPROB = args.trans_prob  # 0.001
    PLOT_REGIONS = args.plot_regions

    # load data
    D0 = util.load_dataframe(intsv_path)
    Matchscores = pd.read_csv(matchscores_path, sep="\t", low_memory=False, header=0, index_col=0)

    # adaptive threshold for group sizes
    minmatchscore = np.median(Matchscores.median()) * SAMPLE_SCORE_FACTOR
    Matchscores_bools = Matchscores < minmatchscore
    selected_samples = Matchscores_bools[Matchscores_bools.sum() >= MINIMUM_SAMPLE_GROUP].index
    failed_samples = Matchscores_bools.index.difference(selected_samples)
    np.fill_diagonal(Matchscores_bools.values, False)
    # optional for neater code downstream
    Matchscores_bools_selected = Matchscores_bools.loc[selected_samples, selected_samples]
    Matchscores_selected = Matchscores.loc[selected_samples, selected_samples]

    if len(selected_samples) == 0:
        print(
            "ERROR: NO SAMPLES IN ANALYSIS. There were no samples selected to perform CNV calling. Try to pick a lower value for MINIMUM_GROUP_SIZES or a higher value for SAMPLE_SCORE_FACTOR. Aborting."
        )
        raise Exception(
            "No samples selected to smaple group. Try to pick a lower value for MINIMUM_GROUP_SIZES or a higher value for SAMPLE_SCORE_FACTOR."
        )

    for p in [calls_path, z_scores_path, ratio_scores_path]:
        pathlib.Path(p).parent.mkdir(parents=True, exist_ok=True)
    pathlib.Path(analysis_dir).mkdir(parents=True, exist_ok=True)

    # evaluation
    # if len(failed_samples) > 0:
    plt.figure(figsize=(6, 4))
    plt.title("Threshold finding of sample groups")
    x = plt.hist(Matchscores.median().sort_values(), bins=30)
    plt.vlines(minmatchscore, 0, max(x[0]), color="orange", label="threshold")
    plt.xlim(0, 1)
    plt.ylabel("number of samples")
    plt.xlabel("median matchscore")
    plt.legend()
    plt.savefig(pathlib.Path(analysis_dir) / "ANALYSIS_group_sizes_threshold.pdf", format="pdf")

    plt.figure(figsize=(6, 4))
    plt.title("Sample group sizes and cutoff")
    x = plt.hist(
        Matchscores_bools.sum().sort_values(),
        bins=int(len(selected_samples) / 3) + 1,
        range=(0, len(selected_samples)),
    )
    plt.vlines(MINIMUM_SAMPLE_GROUP, 0, max(x[0]), color="darkred", label="cutoff")
    plt.xlim(0, len(selected_samples))
    plt.ylabel("number of samples")
    plt.xlabel("group size")
    plt.legend()
    plt.savefig(pathlib.Path(analysis_dir) / "ANALYSIS_group_sizes_cutoff.pdf", format="pdf")

    util.print_clustermap(
        Matchscores_selected,
        path=pathlib.Path(analysis_dir) / "ANALYSIS_clustermap_filtered.pdf",
        title="clustered heat map of selected sample distances",
    )

    # separate chr subsets
    DS0 = util.filter_by_coverage(D0[selected_samples])
    INDEX = DS0.index.copy()
    SAMPLES = DS0.columns.copy()
    DS1 = util.normalize_within_sample(DS0)
    DS1.index = range(len(INDEX))
    I = np.array([(not (s.startswith("X") or s.startswith("Y"))) for s in INDEX])
    DA = (DS1.to_numpy()[I], I)
    I = np.array([s.startswith("X") for s in INDEX])
    DX = (DS1.to_numpy()[I], I)
    I = np.array([s.startswith("Y") for s in INDEX])
    DY = (DS1.to_numpy()[I], I)

    # define data containers
    n, m = len(DS1.columns), len(DS1.index)
    z_scores_scaled = np.zeros((n, m))
    # e_scores        = np.zeros((n,m))
    ratio_scores = 1 + np.zeros((n, m))
    sample_scores = np.zeros(n)

    # parallelized execution
    def collect_result(result):
        sample_id = result[0]
        index = result[1]
        nonlocal sample_scores
        nonlocal ratio_scores
        nonlocal z_scores_scaled
        # E_scores_df.loc[result[1].index,sample]        = result[1]
        sample_scores[sample_id] = result[2] * (sum(index) / len(INDEX))
        ratio_scores[sample_id, index] = result[3]
        z_scores_scaled[sample_id, index] = result[4]

    #  parallelized autosomal calling
    pool = mp.Pool(CORES)
    print("multi threaded calling autosomal targets")
    for i in range(len(selected_samples)):
        pool.apply_async(
            util.calling_cnv,
            args=(
                i,
                DA[0],
                Matchscores_bools_selected,
                DA[1],
                EXPECTED_CNV_RATE,
                ZSCALE,
            ),
            callback=collect_result,
        )
    pool.close()
    pool.join()

    #  calculate expected X-cov ratio and prepare match scores
    if sum(DX[1]) > 0 and np.max(DX[0].mean(axis=0)) / np.min(DX[0].mean(axis=0)) > 1.25:
        kmeans = KMeans(n_clusters=2).fit(DX[0].mean(axis=0).reshape(-1, 1))
        Mx = Matchscores_bools_selected.copy()
        for i, x in enumerate(kmeans.labels_):
            sample = Matchscores_bools_selected.index[i]
            if x:
                Mx[sample] &= kmeans.labels_ == x
            else:
                Mx[sample] &= (1 - kmeans.labels_) == x
        Mx &= Mx.T
        sample_scores_buffer = sample_scores.copy()
        #  --- parallelized X-chromosome calling ---  #
        print("multi threaded calling on X-chromosome")
        pool = mp.Pool(CORES)
        for i in range(len(selected_samples)):
            if sum(Mx.iloc[i]) >= MINIMUM_SAMPLE_GROUP:
                pool.apply_async(
                    util.calling_cnv,
                    args=(
                        i,
                        DX[0],
                        Mx,
                        DX[1],
                        EXPECTED_CNV_RATE,
                        ZSCALE,
                    ),
                    callback=collect_result,
                )
        pool.close()
        pool.join()
        sample_scores += sample_scores_buffer

    #  calculate expected Y-cov ratio and prepare match scores
    if sum(DY[1]) > 0 and np.max(DY[0].mean(axis=0)) / np.min(DY[0].mean(axis=0)) > 1.25:
        kmeans = KMeans(n_clusters=2).fit(DY[0].mean(axis=0).reshape(-1, 1))
        My = Matchscores_bools_selected.copy()
        for i, y in enumerate(kmeans.labels_):
            sample = Matchscores_bools_selected.index[i]
            My[sample] &= kmeans.labels_ == y
        My &= My.T
        sample_scores_buffer = sample_scores.copy()
        #  --- parallelized Y-chromosome calling ---  #
        print("multi threaded calling on Y-chromosome")
        pool = mp.Pool(CORES)
        for i in range(len(selected_samples)):
            if sum(My.iloc[i]) >= MINIMUM_SAMPLE_GROUP:
                pool.apply_async(
                    util.calling_cnv,
                    args=(
                        i,
                        DY[0],
                        My,
                        DY[1],
                        EXPECTED_CNV_RATE,
                        ZSCALE,
                    ),
                    callback=collect_result,
                )
        pool.close()
        pool.join()
        sample_scores += sample_scores_buffer

    # extract single exon CNVs
    print("compute HMM on z-scores")
    # HMM guided CNV candidates
    # =========================================================================
    # HMM PARAMETRIZATION
    mean_del = np.std(z_scores_scaled.flatten(), ddof=1) * (-3.0)
    mean_dup = np.std(z_scores_scaled.flatten(), ddof=1) * 4.0
    mean_median = np.median(z_scores_scaled.flatten())
    # HMM guided CNV candidates
    # =========================================================================
    means = np.array([[mean_del], [mean_median], [mean_dup]])
    p_trans = TRANSPROB
    p_stay = 1.0 - (2.0 * TRANSPROB)
    transitionprobs = np.array(
        [[p_stay, p_trans, p_trans], [p_trans, p_stay, p_trans], [p_trans, p_trans, p_stay]]
    )
    # HMM = np.array([util.hmm_scores(sample, probs, transitionprobs) for sample in z_scores_scaled])
    # =========================================================================
    pool = mp.Pool(CORES)
    HMM = np.array(
        [
            pool.apply(util.hmm_scores, args=([sample, means, transitionprobs]))
            for sample in z_scores_scaled
        ]
    )
    pool.close()
    # =========================================================================

    print("iterate HMM-labels and condense CNVs")
    # extract CNVs from HMM
    cnv_carrying_indexes = np.array(range(z_scores_scaled.shape[0]))[
        [sum((s != 1)) > 0 for s in HMM]
    ]
    pool = mp.Pool(CORES)
    CNVs = list(
        zip(
            cnv_carrying_indexes,
            [
                pool.apply(
                    util.merge_score_cnvs,
                    args=(
                        [
                            HMM[i, :],
                            z_scores_scaled[i, :],
                            ratio_scores[i, :],
                            INDEX,
                            DEL_CUTOFF,
                            DUP_CUTOFF,
                        ]
                    ),
                )
                for i in cnv_carrying_indexes
            ],
        )
    )
    pool.close()

    # format and save results
    RD = pd.DataFrame(
        [
            [SAMPLES[sample_cnv[0]], *cnv.to_list(), sample_scores[CNVs[i][0]]]
            for i, sample_cnv in enumerate(CNVs)
            for cnv in sample_cnv[1]
        ],
        columns=[
            "sample",
            "chr",
            "start",
            "end",
            "gene",
            "aberration",
            "size",
            "score",
            "ratio",
            "sample_score",
        ],
    ).sort_values(by="score", ascending=False)

    # extract single exon CNVs
    RZ = pd.DataFrame(z_scores_scaled, columns=INDEX, index=SAMPLES).T
    RR = pd.DataFrame(ratio_scores, columns=INDEX, index=SAMPLES).T
    S = pd.DataFrame(sample_scores, index=SAMPLES).T
    SINGLE_EXONS = pd.concat(
        [
            pd.DataFrame(
                [
                    [
                        RZ.columns[c],
                        *RZ.index[i].split("_"),
                        "DEL",
                        1,
                        RZ.iloc[i, c],
                        float(S[RZ.columns[c]]),
                        RR.iloc[i, c],
                    ]
                    for i, c in zip(*np.where((RZ < -3.5) & (RR < DEL_CUTOFF)))
                ],
                columns=[
                    "sample",
                    "chr",
                    "start",
                    "end",
                    "gene",
                    "aberration",
                    "size",
                    "score",
                    "sample_score",
                    "ratio",
                ],
            ),
            pd.DataFrame(
                [
                    [
                        RZ.columns[c],
                        *RZ.index[i].split("_"),
                        "DUP",
                        1,
                        RZ.iloc[i, c],
                        float(S[RZ.columns[c]]),
                        RR.iloc[i, c],
                    ]
                    for i, c in zip(*np.where((RZ > 4.5) & (RR > DUP_CUTOFF)))
                ],
                columns=[
                    "sample",
                    "chr",
                    "start",
                    "end",
                    "gene",
                    "aberration",
                    "size",
                    "score",
                    "sample_score",
                    "ratio",
                ],
            ),
        ]
    )
    print("extract single exon CNVs")
    BIG_HITS = ca.flatten_list(
        [
            l.get_hits()
            for l in ca.load_cnvs_from_df(
                RD[
                    [
                        "sample",
                        "chr",
                        "start",
                        "end",
                        "gene",
                        "aberration",
                        "score",
                        "size",
                        "ratio",
                    ]
                ]
            )
        ]
    )
    SINGLE_HITS = ca.flatten_list(
        [
            l.get_hits()
            for l in ca.load_cnvs_from_df(
                SINGLE_EXONS[
                    [
                        "sample",
                        "chr",
                        "start",
                        "end",
                        "gene",
                        "aberration",
                        "score",
                        "size",
                        "ratio",
                    ]
                ]
            )
        ]
    )
    FS = pd.DataFrame(Matchscores_bools.loc[failed_samples].sum(axis=1), columns=["group_size"])
    FS["median_sample_score"] = Matchscores.loc[failed_samples].median(axis=1)
    x_cols = [
        "sample",
        "chr",
        "start",
        "end",
        "gene",
        "aberration",
        "score",
        "size",
        "ratio",
        "sample_score",
    ]
    X = pd.DataFrame(
        [[*h.to_list(), *S[h.to_list()[0]]] for h in ca.hitsA_not_in_hitsB(SINGLE_HITS, BIG_HITS)],
        columns=x_cols,
    )

    FINAL = pd.concat([RD.loc[:, X.columns], X])
    FINAL["score"] = FINAL["score"].astype(float).apply(lambda x: float(abs(x)))
    FINAL = FINAL.sort_values(by="score", ascending=False)
    FINAL.index = list(range(FINAL.shape[0]))

    print("saving results...")
    FINAL.to_csv(calls_path, sep="\t", index=False)
    RZ.to_csv(z_scores_path, sep="\t")
    RR.to_csv(ratio_scores_path, sep="\t")
    S.T.to_csv(pathlib.Path(analysis_dir) / "sample_scores.tsv", sep="\t")
    Matchscores_bools.to_csv(pathlib.Path(analysis_dir) / "samplegroups.tsv", sep="\t")
    FS.to_csv(pathlib.Path(analysis_dir) / "failed_samples.tsv", sep="\t")

    if PLOT_REGIONS and FINAL.shape[0] > 0:
        print("plotting all called CNVs with sample groups...")
        factor = 0.08
        final_regions = [
            "-".join(FINAL.astype(str).loc[i, ["chr", "start", "end"]]) for i in FINAL.index
        ]
        for i, region in enumerate(final_regions):
            print(f"plotting {i} of %d" % len(final_regions))
            sample = FINAL["sample"].iloc[i]
            r = util.select_region(region, RR, sample, Matchscores_bools_selected, buffer=2).clip(
                0, 2
            )
            plt.figure(figsize=(factor * len(r.columns), factor * len(r.index)))
            plt.pcolor(r)
            plt.title(f"{sample} - {region}")
            plt.ylabel("targets")
            plt.xlabel("samples")
            plt.savefig(
                pathlib.Path(analysis_dir) / str(sample + "_" + region + ".pdf"), format="pdf"
            )
            plt.close()

    print("done!")
    return RD, RZ, RR



#!/usr/bin/env python
# coding: utf-8

# In[1]:


import os


# In[2]:


overlap = 0.8
baselap = 0


# In[3]:


def set_overlap(val):
    if val <= 1 and val >= 0:
        global overlap
        overlap = val


# In[4]:


def do_overlap(t1, t2):
    global overlap
    if t1.chr != t2.chr:
        return False
    x = min(t1.end, t2.end) - max(t1.start, t2.start)
    if x <= 0:
        return False
    y = min(t1.end, t2.end) - min(t1.start, t2.start)
    return x / y >= overlap


class Target:
    def __init__(self, chr, start, end, gene):
        self.chr = str(chr)
        self.start = int(start)
        self.end = int(end)
        self.gene = gene

    def __str__(self):
        return "\t".join(map(str, [self.chr, self.start, self.end, self.gene]))

    def __hash__(self):
        return hash(str(self))

    def __ne__(self, other):
        return not self == other

    def __eq__(self, other):
        return do_overlap(self, other)

    def __lt__(self, other):
        if self.chr == other.chr:
            return (
                self.start + (self.end - self.start) / 2
                < other.start + (other.end - other.start) / 2
            )
        elif self.chr.isdigit() and other.chr.isdigit():
            return self.chr < other.chr
        elif not self.chr.isdigit() and not other.chr.isdigit():
            return self.chr < other.chr
        elif self.chr.isdigit() and not other.chr.isdigit():
            return True
        elif not self.chr.isdigit() and other.chr.isdigit():
            return False

    def __gt__(self, other):
        return self == other or not (self < other)

    def __le__(self, other):
        return self < other or self == other

    def __ge__(self, other):
        return self > other or self == other

    def to_list(self):
        return [str(self.chr), int(self.start), int(self.end), str(self.gene)]

    def get_size(self):
        return self.end - self.start


# In[5]:


class CNV:
    def __init__(self, chr, start, end, gene, abb, score=1.0, size=None, ratio=1.0):
        Target.__init__(self, chr, start, end, gene)
        self.abb = abb
        self.size = size
        self.score = float(score)
        self.ratio = float(ratio)

    def __str__(self):
        if self.size:
            return "\t".join(
                map(
                    str,
                    [
                        self.chr,
                        self.start,
                        self.end,
                        self.gene,
                        self.abb,
                        self.score,
                        self.size,
                        self.ratio,
                    ],
                )
            )
        return "\t".join(
            map(str, [self.chr, self.start, self.end, self.gene, self.abb, self.score, self.ratio])
        )

    def __hash__(self):
        return hash("\t".join(str(self).split("\t")[:-2]))

    def __eq__(self, other):
        if type(other) == type(self):
            return do_overlap(self, other) and self.abb == other.abb
        if type(other) == Target:
            return do_overlap(self, other)

    def __ne__(self, other):
        return not self == other

    def __lt__(self, other):
        if self.chr == other.chr:
            return (
                self.start + (self.end - self.start) / 2
                < other.start + (other.end - other.start) / 2
            )
        elif self.chr.isdigit() and other.chr.isdigit():
            return self.chr < other.chr
        elif not self.chr.isdigit() and not other.chr.isdigit():
            return self.chr < other.chr
        elif self.chr.isdigit() and not other.chr.isdigit():
            return True
        elif not self.chr.isdigit() and other.chr.isdigit():
            return False

    def __add__(self, other):
        if self.chr != other.chr:
            raise ValueError("Chromosomes have to be identical.")
        if self.abb != other.abb:
            raise ValueError("Aberrations have to be identical.")
        return CNV(
            self.chr,
            min([self.start, other.start]),
            max([self.end, other.end]),
            self.gene
            if (self.gene in other.gene or other.gene in self.gene)
            else self.gene + "," + other.gene,
            self.abb,
            (self.score * self.size + other.score * other.size) / (self.size + other.size)
            if self.size and other.size
            else 1.0,
            self.size + other.size if self.size and other.size else None,
            (self.ratio * self.size + other.ratio * other.size) / (self.size + other.size)
            if self.size and other.size
            else 1.0,
            self.size + other.size if self.size and other.size else None,
        )

    def to_list(self):
        if self.size:
            return [
                str(self.chr),
                int(self.start),
                int(self.end),
                str(self.gene),
                str(self.abb),
                float(self.score),
                int(self.size),
                float(self.ratio),
            ]
        return [
            str(self.chr),
            int(self.start),
            int(self.end),
            str(self.gene),
            str(self.abb),
            float(self.score),
            float(self.ratio),
        ]

    def get_size(self):
        return self.end - self.start


# In[6]:


class Hit:
    def __init__(self, sample, CNV):
        self.sample = sample
        self.cnv = CNV

    def __str__(self):
        return str(self.sample) + "\t" + str(self.cnv)

    def __hash__(self):
        return hash("\t".join(str(self).split("\t")[:-1]))

    def __eq__(self, other):
        return self.sample == other.sample and self.cnv == other.cnv

    def __ne__(self, other):
        return not self == other

    def to_list(self):
        return [str(self.sample), *self.cnv.to_list()]

    def get_size(self):
        return self.cnv.get_size()


# In[7]:


class Sample:
    def __init__(self, name, CNVs):
        self.name = str(name)
        self.cnvs = sorted(CNVs)

    def __eq__(self, other):
        return self.name == other.name

    def __str__(self):
        return self.name

    def __hash__(self):
        return hash(str(self))

    def has_cnv(self, cnv):
        return cnv in self.cnvs

    def get_hits(self):
        return [Hit(self, cnv) for cnv in self.cnvs]


# In[8]:

# df should be formatted like:
# header = ["sample","chr","start","end","aberration","score","size"]
def load_cnvs_from_df(df):
    samples = []
    header = ["sample", "chr", "start", "end", "gene", "aberration", "score", "size", "ratio"]
    df.columns = header
    df.index = range(len(df.index))
    for sample in set(df["sample"]):
        sub_df = df[df["sample"] == sample]
        cnvs = []
        for target in sub_df.index:
            cnvs.append(CNV(*sub_df.loc[target, header[1:]]))
        samples.append(Sample(sample, cnvs))
    return samples


# In[10]:


def intersect_hits(hits):
    if len(hits) > 1:
        R = []
        for h1 in hits[0]:
            if h1 in hits[1]:
                R.append(h1)
        return intersect_hits([list(set(R)), *hits[2:]])
    else:
        return hits[0]


# In[9]:


def dissect_hits(hits1, hits2):
    R = []
    for h1 in hits1:
        if not h1 in hits2:
            R.append(h1)
    return list(set(R))


def union_hits(hits1, hits2):
    return set(hits1).union(set(hits2))


# In[11]:


def find_samples(hit, hits):
    X = []
    for a in hits:
        if hit.cnv == a.cnv and hit.sample != a.sample:
            X.append(str(a.sample))
    return X


# In[12]:


def read_sample(file, sep="\t"):
    cnvs = []
    with open(file) as f:
        for line in f:
            if sum([s.isdigit() for s in line.split(sep)]) == 0:
                continue
            cnvs.append(CNV(*(line.split(sep))))
    return Sample(os.path.basename(file).split(".")[0], cnvs)


# In[13]:


def overlapping(c1, c2):
    return c1.chr == c2.chr and (min([c1.end, c2.end]) - (max([c2.start, c1.start])) > 0)


def is_hit_unique(hit, hits):
    for h in hits:
        if overlapping(hit.cnv, h.cnv) and hit.sample.name == h.sample.name:
            return False
    return True


def hitsA_not_in_hitsB(hitsA, hitsB):
    R = []
    for h in hitsA:
        if is_hit_unique(h, hitsB):
            R.append(h)
    return R


# bed needs to be formatted: chr, start, end, gene
def num_targets(bed, target, tolerance=10):
    return bed[
        (bed["chr"].astype(str) == str(target.chr))
        & (bed["start"] + tolerance >= int(target.start))
        & (bed["end"] - tolerance <= int(target.end))
    ].shape[0]


def distance(t0, t1):
    if t0.chr == t1.chr:
        return abs((t0.end + t1.start) - (t1.end - t1.start))
    else:
        return 0


# In[10]:


def flatten_list(L):
    return [val for l in L for val in l]
